# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import TYPE_CHECKING, Optional

from xdrlib3 import Packer, Unpacker

from .base import Boolean
from .sc_val_type import SCValType

if TYPE_CHECKING:
    from .duration import Duration
    from .int32 import Int32
    from .int64 import Int64
    from .int128_parts import Int128Parts
    from .int256_parts import Int256Parts
    from .sc_address import SCAddress
    from .sc_bytes import SCBytes
    from .sc_contract_instance import SCContractInstance
    from .sc_error import SCError
    from .sc_map import SCMap
    from .sc_nonce_key import SCNonceKey
    from .sc_string import SCString
    from .sc_symbol import SCSymbol
    from .sc_vec import SCVec
    from .time_point import TimePoint
    from .u_int128_parts import UInt128Parts
    from .u_int256_parts import UInt256Parts
    from .uint32 import Uint32
    from .uint64 import Uint64
__all__ = ["SCVal"]


class SCVal:
    """
    XDR Source Code::

        union SCVal switch (SCValType type)
        {

        case SCV_BOOL:
            bool b;
        case SCV_VOID:
            void;
        case SCV_ERROR:
            SCError error;

        case SCV_U32:
            uint32 u32;
        case SCV_I32:
            int32 i32;

        case SCV_U64:
            uint64 u64;
        case SCV_I64:
            int64 i64;
        case SCV_TIMEPOINT:
            TimePoint timepoint;
        case SCV_DURATION:
            Duration duration;

        case SCV_U128:
            UInt128Parts u128;
        case SCV_I128:
            Int128Parts i128;

        case SCV_U256:
            UInt256Parts u256;
        case SCV_I256:
            Int256Parts i256;

        case SCV_BYTES:
            SCBytes bytes;
        case SCV_STRING:
            SCString str;
        case SCV_SYMBOL:
            SCSymbol sym;

        // Vec and Map are recursive so need to live
        // behind an option, due to xdrpp limitations.
        case SCV_VEC:
            SCVec *vec;
        case SCV_MAP:
            SCMap *map;

        case SCV_ADDRESS:
            SCAddress address;

        // Special SCVals reserved for system-constructed contract-data
        // ledger keys, not generally usable elsewhere.
        case SCV_LEDGER_KEY_CONTRACT_INSTANCE:
            void;
        case SCV_LEDGER_KEY_NONCE:
            SCNonceKey nonce_key;

        case SCV_CONTRACT_INSTANCE:
            SCContractInstance instance;
        };
    """

    def __init__(
        self,
        type: SCValType,
        b: bool = None,
        error: SCError = None,
        u32: Uint32 = None,
        i32: Int32 = None,
        u64: Uint64 = None,
        i64: Int64 = None,
        timepoint: TimePoint = None,
        duration: Duration = None,
        u128: UInt128Parts = None,
        i128: Int128Parts = None,
        u256: UInt256Parts = None,
        i256: Int256Parts = None,
        bytes: SCBytes = None,
        str: SCString = None,
        sym: SCSymbol = None,
        vec: Optional[SCVec] = None,
        map: Optional[SCMap] = None,
        address: SCAddress = None,
        nonce_key: SCNonceKey = None,
        instance: SCContractInstance = None,
    ) -> None:
        self.type = type
        self.b = b
        self.error = error
        self.u32 = u32
        self.i32 = i32
        self.u64 = u64
        self.i64 = i64
        self.timepoint = timepoint
        self.duration = duration
        self.u128 = u128
        self.i128 = i128
        self.u256 = u256
        self.i256 = i256
        self.bytes = bytes
        self.str = str
        self.sym = sym
        self.vec = vec
        self.map = map
        self.address = address
        self.nonce_key = nonce_key
        self.instance = instance

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == SCValType.SCV_BOOL:
            if self.b is None:
                raise ValueError("b should not be None.")
            Boolean(self.b).pack(packer)
            return
        if self.type == SCValType.SCV_VOID:
            return
        if self.type == SCValType.SCV_ERROR:
            if self.error is None:
                raise ValueError("error should not be None.")
            self.error.pack(packer)
            return
        if self.type == SCValType.SCV_U32:
            if self.u32 is None:
                raise ValueError("u32 should not be None.")
            self.u32.pack(packer)
            return
        if self.type == SCValType.SCV_I32:
            if self.i32 is None:
                raise ValueError("i32 should not be None.")
            self.i32.pack(packer)
            return
        if self.type == SCValType.SCV_U64:
            if self.u64 is None:
                raise ValueError("u64 should not be None.")
            self.u64.pack(packer)
            return
        if self.type == SCValType.SCV_I64:
            if self.i64 is None:
                raise ValueError("i64 should not be None.")
            self.i64.pack(packer)
            return
        if self.type == SCValType.SCV_TIMEPOINT:
            if self.timepoint is None:
                raise ValueError("timepoint should not be None.")
            self.timepoint.pack(packer)
            return
        if self.type == SCValType.SCV_DURATION:
            if self.duration is None:
                raise ValueError("duration should not be None.")
            self.duration.pack(packer)
            return
        if self.type == SCValType.SCV_U128:
            if self.u128 is None:
                raise ValueError("u128 should not be None.")
            self.u128.pack(packer)
            return
        if self.type == SCValType.SCV_I128:
            if self.i128 is None:
                raise ValueError("i128 should not be None.")
            self.i128.pack(packer)
            return
        if self.type == SCValType.SCV_U256:
            if self.u256 is None:
                raise ValueError("u256 should not be None.")
            self.u256.pack(packer)
            return
        if self.type == SCValType.SCV_I256:
            if self.i256 is None:
                raise ValueError("i256 should not be None.")
            self.i256.pack(packer)
            return
        if self.type == SCValType.SCV_BYTES:
            if self.bytes is None:
                raise ValueError("bytes should not be None.")
            self.bytes.pack(packer)
            return
        if self.type == SCValType.SCV_STRING:
            if self.str is None:
                raise ValueError("str should not be None.")
            self.str.pack(packer)
            return
        if self.type == SCValType.SCV_SYMBOL:
            if self.sym is None:
                raise ValueError("sym should not be None.")
            self.sym.pack(packer)
            return
        if self.type == SCValType.SCV_VEC:
            if self.vec is None:
                packer.pack_uint(0)
            else:
                packer.pack_uint(1)
                if self.vec is None:
                    raise ValueError("vec should not be None.")
                self.vec.pack(packer)
            return
        if self.type == SCValType.SCV_MAP:
            if self.map is None:
                packer.pack_uint(0)
            else:
                packer.pack_uint(1)
                if self.map is None:
                    raise ValueError("map should not be None.")
                self.map.pack(packer)
            return
        if self.type == SCValType.SCV_ADDRESS:
            if self.address is None:
                raise ValueError("address should not be None.")
            self.address.pack(packer)
            return
        if self.type == SCValType.SCV_LEDGER_KEY_CONTRACT_INSTANCE:
            return
        if self.type == SCValType.SCV_LEDGER_KEY_NONCE:
            if self.nonce_key is None:
                raise ValueError("nonce_key should not be None.")
            self.nonce_key.pack(packer)
            return
        if self.type == SCValType.SCV_CONTRACT_INSTANCE:
            if self.instance is None:
                raise ValueError("instance should not be None.")
            self.instance.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCVal:
        type = SCValType.unpack(unpacker)
        if type == SCValType.SCV_BOOL:
            b = Boolean.unpack(unpacker)
            return cls(type=type, b=b)
        if type == SCValType.SCV_VOID:
            return cls(type=type)
        if type == SCValType.SCV_ERROR:
            from .sc_error import SCError

            error = SCError.unpack(unpacker)
            return cls(type=type, error=error)
        if type == SCValType.SCV_U32:
            from .uint32 import Uint32

            u32 = Uint32.unpack(unpacker)
            return cls(type=type, u32=u32)
        if type == SCValType.SCV_I32:
            from .int32 import Int32

            i32 = Int32.unpack(unpacker)
            return cls(type=type, i32=i32)
        if type == SCValType.SCV_U64:
            from .uint64 import Uint64

            u64 = Uint64.unpack(unpacker)
            return cls(type=type, u64=u64)
        if type == SCValType.SCV_I64:
            from .int64 import Int64

            i64 = Int64.unpack(unpacker)
            return cls(type=type, i64=i64)
        if type == SCValType.SCV_TIMEPOINT:
            from .time_point import TimePoint

            timepoint = TimePoint.unpack(unpacker)
            return cls(type=type, timepoint=timepoint)
        if type == SCValType.SCV_DURATION:
            from .duration import Duration

            duration = Duration.unpack(unpacker)
            return cls(type=type, duration=duration)
        if type == SCValType.SCV_U128:
            from .u_int128_parts import UInt128Parts

            u128 = UInt128Parts.unpack(unpacker)
            return cls(type=type, u128=u128)
        if type == SCValType.SCV_I128:
            from .int128_parts import Int128Parts

            i128 = Int128Parts.unpack(unpacker)
            return cls(type=type, i128=i128)
        if type == SCValType.SCV_U256:
            from .u_int256_parts import UInt256Parts

            u256 = UInt256Parts.unpack(unpacker)
            return cls(type=type, u256=u256)
        if type == SCValType.SCV_I256:
            from .int256_parts import Int256Parts

            i256 = Int256Parts.unpack(unpacker)
            return cls(type=type, i256=i256)
        if type == SCValType.SCV_BYTES:
            from .sc_bytes import SCBytes

            bytes = SCBytes.unpack(unpacker)
            return cls(type=type, bytes=bytes)
        if type == SCValType.SCV_STRING:
            from .sc_string import SCString

            str = SCString.unpack(unpacker)
            return cls(type=type, str=str)
        if type == SCValType.SCV_SYMBOL:
            from .sc_symbol import SCSymbol

            sym = SCSymbol.unpack(unpacker)
            return cls(type=type, sym=sym)
        if type == SCValType.SCV_VEC:
            from .sc_vec import SCVec

            vec = SCVec.unpack(unpacker) if unpacker.unpack_uint() else None
            return cls(type=type, vec=vec)
        if type == SCValType.SCV_MAP:
            from .sc_map import SCMap

            map = SCMap.unpack(unpacker) if unpacker.unpack_uint() else None
            return cls(type=type, map=map)
        if type == SCValType.SCV_ADDRESS:
            from .sc_address import SCAddress

            address = SCAddress.unpack(unpacker)
            return cls(type=type, address=address)
        if type == SCValType.SCV_LEDGER_KEY_CONTRACT_INSTANCE:
            return cls(type=type)
        if type == SCValType.SCV_LEDGER_KEY_NONCE:
            from .sc_nonce_key import SCNonceKey

            nonce_key = SCNonceKey.unpack(unpacker)
            return cls(type=type, nonce_key=nonce_key)
        if type == SCValType.SCV_CONTRACT_INSTANCE:
            from .sc_contract_instance import SCContractInstance

            instance = SCContractInstance.unpack(unpacker)
            return cls(type=type, instance=instance)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCVal:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCVal:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.b,
                self.error,
                self.u32,
                self.i32,
                self.u64,
                self.i64,
                self.timepoint,
                self.duration,
                self.u128,
                self.i128,
                self.u256,
                self.i256,
                self.bytes,
                self.str,
                self.sym,
                self.vec,
                self.map,
                self.address,
                self.nonce_key,
                self.instance,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.b == other.b
            and self.error == other.error
            and self.u32 == other.u32
            and self.i32 == other.i32
            and self.u64 == other.u64
            and self.i64 == other.i64
            and self.timepoint == other.timepoint
            and self.duration == other.duration
            and self.u128 == other.u128
            and self.i128 == other.i128
            and self.u256 == other.u256
            and self.i256 == other.i256
            and self.bytes == other.bytes
            and self.str == other.str
            and self.sym == other.sym
            and self.vec == other.vec
            and self.map == other.map
            and self.address == other.address
            and self.nonce_key == other.nonce_key
            and self.instance == other.instance
        )

    def __repr__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(f"b={self.b}") if self.b is not None else None
        out.append(f"error={self.error}") if self.error is not None else None
        out.append(f"u32={self.u32}") if self.u32 is not None else None
        out.append(f"i32={self.i32}") if self.i32 is not None else None
        out.append(f"u64={self.u64}") if self.u64 is not None else None
        out.append(f"i64={self.i64}") if self.i64 is not None else None
        (
            out.append(f"timepoint={self.timepoint}")
            if self.timepoint is not None
            else None
        )
        out.append(f"duration={self.duration}") if self.duration is not None else None
        out.append(f"u128={self.u128}") if self.u128 is not None else None
        out.append(f"i128={self.i128}") if self.i128 is not None else None
        out.append(f"u256={self.u256}") if self.u256 is not None else None
        out.append(f"i256={self.i256}") if self.i256 is not None else None
        out.append(f"bytes={self.bytes}") if self.bytes is not None else None
        out.append(f"str={self.str}") if self.str is not None else None
        out.append(f"sym={self.sym}") if self.sym is not None else None
        out.append(f"vec={self.vec}") if self.vec is not None else None
        out.append(f"map={self.map}") if self.map is not None else None
        out.append(f"address={self.address}") if self.address is not None else None
        (
            out.append(f"nonce_key={self.nonce_key}")
            if self.nonce_key is not None
            else None
        )
        out.append(f"instance={self.instance}") if self.instance is not None else None
        return f"<SCVal [{', '.join(out)}]>"



from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..keypair import Keypair
from ..muxed_account import MuxedAccount
from ..strkey import StrKey
from ..utils import raise_if_not_valid_ed25519_public_key
from .operation import Operation

__all__ = ["BeginSponsoringFutureReserves"]


class BeginSponsoringFutureReserves(Operation):
    """The :class:`BeginSponsoringFutureReserves` object, which represents a BeginSponsoringFutureReserves
    operation on Stellar's network.

    Establishes the is-sponsoring-future-reserves-for relationship between the source account and sponsoredID.
    See `Sponsored Reserves <https://developers.stellar.org/docs/glossary/sponsored-reserves/>`_ for more information.

    Threshold: Medium

    See `Begin Sponsoring Future Reserves <https://developers.stellar.org/docs/start/list-of-operations/#begin-sponsoring-future-reserves>`_ for more information.

    :param sponsored_id: The sponsored account id.
    :param source: The source account for the operation. Defaults to the transaction's source account.
    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.BEGIN_SPONSORING_FUTURE_RESERVES
    )

    def __init__(
        self, sponsored_id: str, source: Optional[Union[MuxedAccount, str]] = None
    ) -> None:
        super().__init__(source)
        self.sponsored_id: str = sponsored_id
        raise_if_not_valid_ed25519_public_key(self.sponsored_id, "sponsored_id")

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        sponsored_id = Keypair.from_public_key(self.sponsored_id).xdr_account_id()
        begin_sponsoring_future_reserves_op = (
            stellar_xdr.BeginSponsoringFutureReservesOp(sponsored_id=sponsored_id)
        )
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE,
            begin_sponsoring_future_reserves_op=begin_sponsoring_future_reserves_op,
        )
        return body

    @classmethod
    def from_xdr_object(
        cls, xdr_object: stellar_xdr.Operation
    ) -> "BeginSponsoringFutureReserves":
        """Creates a :class:`BeginSponsoringFutureReserves` object from an XDR Operation
        object.
        """
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.begin_sponsoring_future_reserves_op is not None
        assert (
            xdr_object.body.begin_sponsoring_future_reserves_op.sponsored_id.account_id.ed25519
            is not None
        )
        sponsored_id = StrKey.encode_ed25519_public_key(
            xdr_object.body.begin_sponsoring_future_reserves_op.sponsored_id.account_id.ed25519.uint256
        )
        op = cls(source=source, sponsored_id=sponsored_id)
        return op

    def __repr__(self):
        return f"<BeginSponsoringFutureReserves [sponsored_id={self.sponsored_id}, source={self.source}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .decorated_signature import DecoratedSignature
from .transaction import Transaction

__all__ = ["TransactionV1Envelope"]


class TransactionV1Envelope:
    """
    XDR Source Code::

        struct TransactionV1Envelope
        {
            Transaction tx;
            /* Each decorated signature is a signature over the SHA256 hash of
             * a TransactionSignaturePayload */
            DecoratedSignature signatures<20>;
        };
    """

    def __init__(
        self,
        tx: Transaction,
        signatures: List[DecoratedSignature],
    ) -> None:
        _expect_max_length = 20
        if signatures and len(signatures) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `signatures` should be {_expect_max_length}, but got {len(signatures)}."
            )
        self.tx = tx
        self.signatures = signatures

    def pack(self, packer: Packer) -> None:
        self.tx.pack(packer)
        packer.pack_uint(len(self.signatures))
        for signatures_item in self.signatures:
            signatures_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TransactionV1Envelope:
        tx = Transaction.unpack(unpacker)
        length = unpacker.unpack_uint()
        signatures = []
        for _ in range(length):
            signatures.append(DecoratedSignature.unpack(unpacker))
        return cls(
            tx=tx,
            signatures=signatures,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TransactionV1Envelope:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TransactionV1Envelope:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.tx,
                self.signatures,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.tx == other.tx and self.signatures == other.signatures

    def __repr__(self):
        out = [
            f"tx={self.tx}",
            f"signatures={self.signatures}",
        ]
        return f"<TransactionV1Envelope [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .bucket_metadata import BucketMetadata
from .cold_archive_archived_leaf import ColdArchiveArchivedLeaf
from .cold_archive_boundary_leaf import ColdArchiveBoundaryLeaf
from .cold_archive_bucket_entry_type import ColdArchiveBucketEntryType
from .cold_archive_deleted_leaf import ColdArchiveDeletedLeaf
from .cold_archive_hash_entry import ColdArchiveHashEntry

__all__ = ["ColdArchiveBucketEntry"]


class ColdArchiveBucketEntry:
    """
    XDR Source Code::

        union ColdArchiveBucketEntry switch (ColdArchiveBucketEntryType type)
        {
        case COLD_ARCHIVE_METAENTRY:
            BucketMetadata metaEntry;
        case COLD_ARCHIVE_ARCHIVED_LEAF:
            ColdArchiveArchivedLeaf archivedLeaf;
        case COLD_ARCHIVE_DELETED_LEAF:
            ColdArchiveDeletedLeaf deletedLeaf;
        case COLD_ARCHIVE_BOUNDARY_LEAF:
            ColdArchiveBoundaryLeaf boundaryLeaf;
        case COLD_ARCHIVE_HASH:
            ColdArchiveHashEntry hashEntry;
        };
    """

    def __init__(
        self,
        type: ColdArchiveBucketEntryType,
        meta_entry: BucketMetadata = None,
        archived_leaf: ColdArchiveArchivedLeaf = None,
        deleted_leaf: ColdArchiveDeletedLeaf = None,
        boundary_leaf: ColdArchiveBoundaryLeaf = None,
        hash_entry: ColdArchiveHashEntry = None,
    ) -> None:
        self.type = type
        self.meta_entry = meta_entry
        self.archived_leaf = archived_leaf
        self.deleted_leaf = deleted_leaf
        self.boundary_leaf = boundary_leaf
        self.hash_entry = hash_entry

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == ColdArchiveBucketEntryType.COLD_ARCHIVE_METAENTRY:
            if self.meta_entry is None:
                raise ValueError("meta_entry should not be None.")
            self.meta_entry.pack(packer)
            return
        if self.type == ColdArchiveBucketEntryType.COLD_ARCHIVE_ARCHIVED_LEAF:
            if self.archived_leaf is None:
                raise ValueError("archived_leaf should not be None.")
            self.archived_leaf.pack(packer)
            return
        if self.type == ColdArchiveBucketEntryType.COLD_ARCHIVE_DELETED_LEAF:
            if self.deleted_leaf is None:
                raise ValueError("deleted_leaf should not be None.")
            self.deleted_leaf.pack(packer)
            return
        if self.type == ColdArchiveBucketEntryType.COLD_ARCHIVE_BOUNDARY_LEAF:
            if self.boundary_leaf is None:
                raise ValueError("boundary_leaf should not be None.")
            self.boundary_leaf.pack(packer)
            return
        if self.type == ColdArchiveBucketEntryType.COLD_ARCHIVE_HASH:
            if self.hash_entry is None:
                raise ValueError("hash_entry should not be None.")
            self.hash_entry.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ColdArchiveBucketEntry:
        type = ColdArchiveBucketEntryType.unpack(unpacker)
        if type == ColdArchiveBucketEntryType.COLD_ARCHIVE_METAENTRY:
            meta_entry = BucketMetadata.unpack(unpacker)
            return cls(type=type, meta_entry=meta_entry)
        if type == ColdArchiveBucketEntryType.COLD_ARCHIVE_ARCHIVED_LEAF:
            archived_leaf = ColdArchiveArchivedLeaf.unpack(unpacker)
            return cls(type=type, archived_leaf=archived_leaf)
        if type == ColdArchiveBucketEntryType.COLD_ARCHIVE_DELETED_LEAF:
            deleted_leaf = ColdArchiveDeletedLeaf.unpack(unpacker)
            return cls(type=type, deleted_leaf=deleted_leaf)
        if type == ColdArchiveBucketEntryType.COLD_ARCHIVE_BOUNDARY_LEAF:
            boundary_leaf = ColdArchiveBoundaryLeaf.unpack(unpacker)
            return cls(type=type, boundary_leaf=boundary_leaf)
        if type == ColdArchiveBucketEntryType.COLD_ARCHIVE_HASH:
            hash_entry = ColdArchiveHashEntry.unpack(unpacker)
            return cls(type=type, hash_entry=hash_entry)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ColdArchiveBucketEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ColdArchiveBucketEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.meta_entry,
                self.archived_leaf,
                self.deleted_leaf,
                self.boundary_leaf,
                self.hash_entry,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.meta_entry == other.meta_entry
            and self.archived_leaf == other.archived_leaf
            and self.deleted_leaf == other.deleted_leaf
            and self.boundary_leaf == other.boundary_leaf
            and self.hash_entry == other.hash_entry
        )

    def __repr__(self):
        out = []
        out.append(f"type={self.type}")
        (
            out.append(f"meta_entry={self.meta_entry}")
            if self.meta_entry is not None
            else None
        )
        (
            out.append(f"archived_leaf={self.archived_leaf}")
            if self.archived_leaf is not None
            else None
        )
        (
            out.append(f"deleted_leaf={self.deleted_leaf}")
            if self.deleted_leaf is not None
            else None
        )
        (
            out.append(f"boundary_leaf={self.boundary_leaf}")
            if self.boundary_leaf is not None
            else None
        )
        (
            out.append(f"hash_entry={self.hash_entry}")
            if self.hash_entry is not None
            else None
        )
        return f"<ColdArchiveBucketEntry [{', '.join(out)}]>"



from __future__ import annotations

import copy
import dataclasses
import uuid
from typing import TYPE_CHECKING

from . import xdr as stellar_xdr
from .operation import InvokeHostFunction
from .soroban_rpc import *

if TYPE_CHECKING:
    from .transaction_envelope import TransactionEnvelope


@dataclasses.dataclass
class ResourceLeeway:
    """Describes additional resource leeways for transaction simulation."""

    cpu_instructions: int


class Durability(Enum):
    TEMPORARY = "temporary"
    PERSISTENT = "persistent"


def _generate_unique_request_id() -> str:
    return uuid.uuid4().hex


def _assemble_transaction(
    transaction_envelope: TransactionEnvelope,
    simulation: SimulateTransactionResponse,
) -> TransactionEnvelope:
    # TODO: add support for FeeBumpTransactionEnvelope
    if not transaction_envelope.transaction.is_soroban_transaction():
        raise ValueError(
            "Unsupported transaction: must contain exactly one operation of "
            "type RestoreFootprint, InvokeHostFunction or ExtendFootprintTTL"
        )

    min_resource_fee = simulation.min_resource_fee
    assert simulation.transaction_data is not None
    soroban_data = stellar_xdr.SorobanTransactionData.from_xdr(
        simulation.transaction_data
    )
    te = copy.deepcopy(transaction_envelope)
    te.signatures = []

    # Reset fee to the original value, excluding the resource fee
    if te.transaction.soroban_data:
        te.transaction.fee -= te.transaction.soroban_data.resource_fee.int64

    assert min_resource_fee is not None
    te.transaction.fee += min_resource_fee
    te.transaction.soroban_data = soroban_data

    op = te.transaction.operations[0]

    if isinstance(op, InvokeHostFunction):
        if not simulation.results or len(simulation.results) != 1:
            raise ValueError(f"Simulation results invalid: {simulation.results}")

        if not op.auth and simulation.results[0].auth:
            op.auth = [
                stellar_xdr.SorobanAuthorizationEntry.from_xdr(xdr)
                for xdr in simulation.results[0].auth
            ]
    return te



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .uint32 import Uint32

__all__ = ["SCEnvMetaEntryInterfaceVersion"]


class SCEnvMetaEntryInterfaceVersion:
    """
    XDR Source Code::

        struct {
                uint32 protocol;
                uint32 preRelease;
            }
    """

    def __init__(
        self,
        protocol: Uint32,
        pre_release: Uint32,
    ) -> None:
        self.protocol = protocol
        self.pre_release = pre_release

    def pack(self, packer: Packer) -> None:
        self.protocol.pack(packer)
        self.pre_release.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCEnvMetaEntryInterfaceVersion:
        protocol = Uint32.unpack(unpacker)
        pre_release = Uint32.unpack(unpacker)
        return cls(
            protocol=protocol,
            pre_release=pre_release,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCEnvMetaEntryInterfaceVersion:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCEnvMetaEntryInterfaceVersion:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.protocol,
                self.pre_release,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.protocol == other.protocol and self.pre_release == other.pre_release

    def __repr__(self):
        out = [
            f"protocol={self.protocol}",
            f"pre_release={self.pre_release}",
        ]
        return f"<SCEnvMetaEntryInterfaceVersion [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import Opaque
from .ip_addr_type import IPAddrType

__all__ = ["PeerAddressIp"]


class PeerAddressIp:
    """
    XDR Source Code::

        union switch (IPAddrType type)
            {
            case IPv4:
                opaque ipv4[4];
            case IPv6:
                opaque ipv6[16];
            }
    """

    def __init__(
        self,
        type: IPAddrType,
        ipv4: bytes = None,
        ipv6: bytes = None,
    ) -> None:
        self.type = type
        self.ipv4 = ipv4
        self.ipv6 = ipv6

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == IPAddrType.IPv4:
            if self.ipv4 is None:
                raise ValueError("ipv4 should not be None.")
            Opaque(self.ipv4, 4, True).pack(packer)
            return
        if self.type == IPAddrType.IPv6:
            if self.ipv6 is None:
                raise ValueError("ipv6 should not be None.")
            Opaque(self.ipv6, 16, True).pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> PeerAddressIp:
        type = IPAddrType.unpack(unpacker)
        if type == IPAddrType.IPv4:
            ipv4 = Opaque.unpack(unpacker, 4, True)
            return cls(type=type, ipv4=ipv4)
        if type == IPAddrType.IPv6:
            ipv6 = Opaque.unpack(unpacker, 16, True)
            return cls(type=type, ipv6=ipv6)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> PeerAddressIp:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> PeerAddressIp:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.ipv4,
                self.ipv6,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.ipv4 == other.ipv4
            and self.ipv6 == other.ipv6
        )

    def __repr__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(f"ipv4={self.ipv4}") if self.ipv4 is not None else None
        out.append(f"ipv6={self.ipv6}") if self.ipv6 is not None else None
        return f"<PeerAddressIp [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .sc_address import SCAddress
from .sc_symbol import SCSymbol
from .sc_val import SCVal

__all__ = ["InvokeContractArgs"]


class InvokeContractArgs:
    """
    XDR Source Code::

        struct InvokeContractArgs {
            SCAddress contractAddress;
            SCSymbol functionName;
            SCVal args<>;
        };
    """

    def __init__(
        self,
        contract_address: SCAddress,
        function_name: SCSymbol,
        args: List[SCVal],
    ) -> None:
        _expect_max_length = 4294967295
        if args and len(args) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `args` should be {_expect_max_length}, but got {len(args)}."
            )
        self.contract_address = contract_address
        self.function_name = function_name
        self.args = args

    def pack(self, packer: Packer) -> None:
        self.contract_address.pack(packer)
        self.function_name.pack(packer)
        packer.pack_uint(len(self.args))
        for args_item in self.args:
            args_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> InvokeContractArgs:
        contract_address = SCAddress.unpack(unpacker)
        function_name = SCSymbol.unpack(unpacker)
        length = unpacker.unpack_uint()
        args = []
        for _ in range(length):
            args.append(SCVal.unpack(unpacker))
        return cls(
            contract_address=contract_address,
            function_name=function_name,
            args=args,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> InvokeContractArgs:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> InvokeContractArgs:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.contract_address,
                self.function_name,
                self.args,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.contract_address == other.contract_address
            and self.function_name == other.function_name
            and self.args == other.args
        )

    def __repr__(self):
        out = [
            f"contract_address={self.contract_address}",
            f"function_name={self.function_name}",
            f"args={self.args}",
        ]
        return f"<InvokeContractArgs [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .contract_data_durability import ContractDataDurability
from .extension_point import ExtensionPoint
from .sc_address import SCAddress
from .sc_val import SCVal

__all__ = ["ContractDataEntry"]


class ContractDataEntry:
    """
    XDR Source Code::

        struct ContractDataEntry {
            ExtensionPoint ext;

            SCAddress contract;
            SCVal key;
            ContractDataDurability durability;
            SCVal val;
        };
    """

    def __init__(
        self,
        ext: ExtensionPoint,
        contract: SCAddress,
        key: SCVal,
        durability: ContractDataDurability,
        val: SCVal,
    ) -> None:
        self.ext = ext
        self.contract = contract
        self.key = key
        self.durability = durability
        self.val = val

    def pack(self, packer: Packer) -> None:
        self.ext.pack(packer)
        self.contract.pack(packer)
        self.key.pack(packer)
        self.durability.pack(packer)
        self.val.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ContractDataEntry:
        ext = ExtensionPoint.unpack(unpacker)
        contract = SCAddress.unpack(unpacker)
        key = SCVal.unpack(unpacker)
        durability = ContractDataDurability.unpack(unpacker)
        val = SCVal.unpack(unpacker)
        return cls(
            ext=ext,
            contract=contract,
            key=key,
            durability=durability,
            val=val,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ContractDataEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ContractDataEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ext,
                self.contract,
                self.key,
                self.durability,
                self.val,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.ext == other.ext
            and self.contract == other.contract
            and self.key == other.key
            and self.durability == other.durability
            and self.val == other.val
        )

    def __repr__(self):
        out = [
            f"ext={self.ext}",
            f"contract={self.contract}",
            f"key={self.key}",
            f"durability={self.durability}",
            f"val={self.val}",
        ]
        return f"<ContractDataEntry [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .ledger_entry_change import LedgerEntryChange

__all__ = ["LedgerEntryChanges"]


class LedgerEntryChanges:
    """
    XDR Source Code::

        typedef LedgerEntryChange LedgerEntryChanges<>;
    """

    def __init__(self, ledger_entry_changes: List[LedgerEntryChange]) -> None:
        _expect_max_length = 4294967295
        if ledger_entry_changes and len(ledger_entry_changes) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `ledger_entry_changes` should be {_expect_max_length}, but got {len(ledger_entry_changes)}."
            )
        self.ledger_entry_changes = ledger_entry_changes

    def pack(self, packer: Packer) -> None:
        packer.pack_uint(len(self.ledger_entry_changes))
        for ledger_entry_changes_item in self.ledger_entry_changes:
            ledger_entry_changes_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LedgerEntryChanges:
        length = unpacker.unpack_uint()
        ledger_entry_changes = []
        for _ in range(length):
            ledger_entry_changes.append(LedgerEntryChange.unpack(unpacker))
        return cls(ledger_entry_changes)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LedgerEntryChanges:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LedgerEntryChanges:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(self.ledger_entry_changes)

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.ledger_entry_changes == other.ledger_entry_changes

    def __repr__(self):
        return (
            f"<LedgerEntryChanges [ledger_entry_changes={self.ledger_entry_changes}]>"
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["MessageType"]


class MessageType(IntEnum):
    """
    XDR Source Code::

        enum MessageType
        {
            ERROR_MSG = 0,
            AUTH = 2,
            DONT_HAVE = 3,

            GET_PEERS = 4, // gets a list of peers this guy knows about
            PEERS = 5,

            GET_TX_SET = 6, // gets a particular txset by hash
            TX_SET = 7,
            GENERALIZED_TX_SET = 17,

            TRANSACTION = 8, // pass on a tx you have heard about

            // SCP
            GET_SCP_QUORUMSET = 9,
            SCP_QUORUMSET = 10,
            SCP_MESSAGE = 11,
            GET_SCP_STATE = 12,

            // new messages
            HELLO = 13,

            SURVEY_REQUEST = 14,
            SURVEY_RESPONSE = 15,

            SEND_MORE = 16,
            SEND_MORE_EXTENDED = 20,

            FLOOD_ADVERT = 18,
            FLOOD_DEMAND = 19,

            TIME_SLICED_SURVEY_REQUEST = 21,
            TIME_SLICED_SURVEY_RESPONSE = 22,
            TIME_SLICED_SURVEY_START_COLLECTING = 23,
            TIME_SLICED_SURVEY_STOP_COLLECTING = 24
        };
    """

    ERROR_MSG = 0
    AUTH = 2
    DONT_HAVE = 3
    GET_PEERS = 4
    PEERS = 5
    GET_TX_SET = 6
    TX_SET = 7
    GENERALIZED_TX_SET = 17
    TRANSACTION = 8
    GET_SCP_QUORUMSET = 9
    SCP_QUORUMSET = 10
    SCP_MESSAGE = 11
    GET_SCP_STATE = 12
    HELLO = 13
    SURVEY_REQUEST = 14
    SURVEY_RESPONSE = 15
    SEND_MORE = 16
    SEND_MORE_EXTENDED = 20
    FLOOD_ADVERT = 18
    FLOOD_DEMAND = 19
    TIME_SLICED_SURVEY_REQUEST = 21
    TIME_SLICED_SURVEY_RESPONSE = 22
    TIME_SLICED_SURVEY_START_COLLECTING = 23
    TIME_SLICED_SURVEY_STOP_COLLECTING = 24

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> MessageType:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> MessageType:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> MessageType:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["HotArchiveBucketEntryType"]


class HotArchiveBucketEntryType(IntEnum):
    """
    XDR Source Code::

        enum HotArchiveBucketEntryType
        {
            HOT_ARCHIVE_METAENTRY = -1, // Bucket metadata, should come first.
            HOT_ARCHIVE_ARCHIVED = 0,   // Entry is Archived
            HOT_ARCHIVE_LIVE = 1,       // Entry was previously HOT_ARCHIVE_ARCHIVED, or HOT_ARCHIVE_DELETED, but
                                        // has been added back to the live BucketList.
                                        // Does not need to be persisted.
            HOT_ARCHIVE_DELETED = 2     // Entry deleted (Note: must be persisted in archive)
        };
    """

    HOT_ARCHIVE_METAENTRY = -1
    HOT_ARCHIVE_ARCHIVED = 0
    HOT_ARCHIVE_LIVE = 1
    HOT_ARCHIVE_DELETED = 2

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> HotArchiveBucketEntryType:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> HotArchiveBucketEntryType:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> HotArchiveBucketEntryType:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



from typing import Callable, Optional, Sequence, TypeVar, Union

from .. import Account, Asset, Keypair, MuxedAccount, scval
from .. import xdr as stellar_xdr
from ..client.base_sync_client import BaseSyncClient
from ..soroban_server import SorobanServer
from ..transaction_builder import TransactionBuilder
from .assembled_transaction import AssembledTransaction

NULL_ACCOUNT = "GAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAWHF"

__all__ = ["ContractClient"]

T = TypeVar("T")


class ContractClient:
    """A client to interact with Soroban smart contracts.

    This client is a wrapper for :py:class:`TransactionBuilder <stellar_sdk.TransactionBuilder>` and :py:class:`SorobanServer <stellar_sdk.SorobanServer>`.
    If you need more fine-grained control, please consider using them directly.

    :param contract_id: The ID of the Soroban contract.
    :param rpc_url: The URL of the RPC server.
    :param network_passphrase: The network passphrase.
    :param request_client: The request client used to send the request.
    """

    def __init__(
        self,
        contract_id: str,
        rpc_url: str,
        network_passphrase: str,
        request_client: BaseSyncClient = None,
    ):
        self.contract_id = contract_id
        self.rpc_url = rpc_url
        self.network_passphrase = network_passphrase
        self.server = SorobanServer(rpc_url, request_client)

    def invoke(
        self,
        function_name: str,
        parameters: Sequence[stellar_xdr.SCVal] = None,
        source: Union[str, MuxedAccount] = NULL_ACCOUNT,
        signer: Keypair = None,
        parse_result_xdr_fn: Optional[Callable[[stellar_xdr.SCVal], T]] = None,
        base_fee: int = 100,
        transaction_timeout: int = 300,
        submit_timeout: int = 120,
        simulate: bool = True,
        restore: bool = True,
    ) -> AssembledTransaction[T]:
        """Build an :py:class:`AssembledTransaction <stellar_sdk.contract.AssembledTransaction>` to invoke a contract function.

        :param function_name: The name of the function to invoke.
        :param parameters: The parameters to pass to the function.
        :param source: The source account for the transaction.
        :param signer: The signer for the transaction.
        :param parse_result_xdr_fn: The function to parse the result XDR returned by the contract function, keep the result as :py:class:`SCVal <stellar_sdk.xdr.SCVal>` if not provided.
        :param base_fee: The base fee for the transaction.
        :param transaction_timeout: The timeout for the transaction.
        :param submit_timeout: The timeout for submitting the transaction.
        :param simulate: Whether to simulate the transaction.
        :param restore: Whether to restore the transaction, only valid when simulate is True, and the signer is provided.
        :return:
        """
        builder = (
            TransactionBuilder(
                Account(source, 0), self.network_passphrase, base_fee=base_fee
            )
            .append_invoke_contract_function_op(
                self.contract_id, function_name, parameters
            )
            .set_timeout(transaction_timeout)
        )
        assembled = AssembledTransaction(
            builder, self.server, signer, parse_result_xdr_fn, submit_timeout
        )
        if simulate:
            assembled = assembled.simulate(restore)
        return assembled

    @staticmethod
    def upload_contract_wasm(
        contract: Union[bytes, str],
        source: Union[str, MuxedAccount],
        signer: Keypair,
        soroban_server: SorobanServer,
        network_passphrase: str = None,
        base_fee: int = 100,
        transaction_timeout: int = 300,
        submit_timeout: int = 120,
    ) -> bytes:
        """Upload a contract wasm.

        :param contract: The contract wasm.
        :param source: The source account for the transaction.
        :param signer: The signer for the transaction.
        :param soroban_server: The Soroban server.
        :param network_passphrase: The network passphrase, default to the network of the Soroban server.
        :param base_fee: The base fee for the transaction.
        :param transaction_timeout: The timeout for the transaction.
        :param submit_timeout: The timeout for submitting the transaction.
        :return: The wasm ID.
        """
        if network_passphrase is None:
            network_passphrase = soroban_server.get_network().passphrase
        builder = (
            TransactionBuilder(
                Account(source, 0), network_passphrase, base_fee=base_fee
            )
            .append_upload_contract_wasm_op(contract)
            .set_timeout(transaction_timeout)
        )
        wasm_id = (
            AssembledTransaction[bytes](
                builder,
                soroban_server,
                signer,
                lambda v: scval.from_bytes(v),
                submit_timeout,
            )
            .simulate()
            .sign_and_submit(force=True)
        )
        assert isinstance(wasm_id, bytes)
        return wasm_id

    @staticmethod
    def create_contract(
        wasm_id: Union[bytes, str],
        source: Union[str, MuxedAccount],
        signer: Keypair,
        soroban_server: SorobanServer,
        constructor_args: Optional[Sequence[stellar_xdr.SCVal]] = None,
        salt: Optional[bytes] = None,
        network_passphrase: str = None,
        base_fee: int = 100,
        transaction_timeout: int = 300,
        submit_timeout: int = 120,
        restore: bool = True,
    ) -> str:
        """Create a contract.

        :param wasm_id: The wasm ID.
        :param source: The source account for the transaction.
        :param signer: The signer for the transaction.
        :param soroban_server: The Soroban server.
        :param constructor_args: The constructor arguments.
        :param salt: The salt.
        :param network_passphrase: The network passphrase, default to the network of the Soroban server.
        :param base_fee: The base fee for the transaction.
        :param transaction_timeout: The timeout for the transaction.
        :param submit_timeout: The timeout for submitting the transaction.
        :param restore: Whether to restore the transaction.
        :return: The contract ID.
        """
        if network_passphrase is None:
            network_passphrase = soroban_server.get_network().passphrase
        address = source if isinstance(source, str) else source.account_id
        builder = (
            TransactionBuilder(
                Account(source, 0), network_passphrase, base_fee=base_fee
            )
            .append_create_contract_op(wasm_id, address, constructor_args, salt)
            .set_timeout(transaction_timeout)
        )
        contract_id = (
            AssembledTransaction[str](
                builder,
                soroban_server,
                signer,
                lambda v: scval.from_address(v).address,
                submit_timeout,
            )
            .simulate(restore)
            .sign_and_submit(force=True)
        )
        assert isinstance(contract_id, str)
        return contract_id

    @staticmethod
    def create_stellar_asset_contract_from_asset(
        asset: Asset,
        source: Union[str, MuxedAccount],
        signer: Keypair,
        soroban_server: SorobanServer,
        network_passphrase: str = None,
        base_fee: int = 100,
        submit_timeout: int = 120,
    ) -> str:
        """Create a Stellar asset contract from an asset.

        :param asset: The asset.
        :param source: The source account for the transaction.
        :param signer: The signer for the transaction.
        :param soroban_server: The Soroban server.
        :param network_passphrase: The network passphrase, default to the network of the Soroban server.
        :param base_fee: The base fee for the transaction.
        :param submit_timeout: The timeout for submitting the transaction.
        :return: The contract ID.
        """
        if network_passphrase is None:
            network_passphrase = soroban_server.get_network().passphrase
        builder = (
            TransactionBuilder(
                Account(source, 0), network_passphrase, base_fee=base_fee
            )
            .append_create_stellar_asset_contract_from_asset_op(asset, source)
            .add_time_bounds(0, 0)
        )
        contract_id = (
            AssembledTransaction[str](
                builder,
                soroban_server,
                signer,
                lambda v: scval.from_address(v).address,
                submit_timeout,
            )
            .simulate()
            .sign_and_submit(force=True)
        )
        assert isinstance(contract_id, str)
        return contract_id

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.server.close()



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["ManageSellOfferResultCode"]


class ManageSellOfferResultCode(IntEnum):
    """
    XDR Source Code::

        enum ManageSellOfferResultCode
        {
            // codes considered as "success" for the operation
            MANAGE_SELL_OFFER_SUCCESS = 0,

            // codes considered as "failure" for the operation
            MANAGE_SELL_OFFER_MALFORMED = -1, // generated offer would be invalid
            MANAGE_SELL_OFFER_SELL_NO_TRUST =
                -2,                              // no trust line for what we're selling
            MANAGE_SELL_OFFER_BUY_NO_TRUST = -3, // no trust line for what we're buying
            MANAGE_SELL_OFFER_SELL_NOT_AUTHORIZED = -4, // not authorized to sell
            MANAGE_SELL_OFFER_BUY_NOT_AUTHORIZED = -5,  // not authorized to buy
            MANAGE_SELL_OFFER_LINE_FULL = -6, // can't receive more of what it's buying
            MANAGE_SELL_OFFER_UNDERFUNDED = -7, // doesn't hold what it's trying to sell
            MANAGE_SELL_OFFER_CROSS_SELF =
                -8, // would cross an offer from the same user
            MANAGE_SELL_OFFER_SELL_NO_ISSUER = -9, // no issuer for what we're selling
            MANAGE_SELL_OFFER_BUY_NO_ISSUER = -10, // no issuer for what we're buying

            // update errors
            MANAGE_SELL_OFFER_NOT_FOUND =
                -11, // offerID does not match an existing offer

            MANAGE_SELL_OFFER_LOW_RESERVE =
                -12 // not enough funds to create a new Offer
        };
    """

    MANAGE_SELL_OFFER_SUCCESS = 0
    MANAGE_SELL_OFFER_MALFORMED = -1
    MANAGE_SELL_OFFER_SELL_NO_TRUST = -2
    MANAGE_SELL_OFFER_BUY_NO_TRUST = -3
    MANAGE_SELL_OFFER_SELL_NOT_AUTHORIZED = -4
    MANAGE_SELL_OFFER_BUY_NOT_AUTHORIZED = -5
    MANAGE_SELL_OFFER_LINE_FULL = -6
    MANAGE_SELL_OFFER_UNDERFUNDED = -7
    MANAGE_SELL_OFFER_CROSS_SELF = -8
    MANAGE_SELL_OFFER_SELL_NO_ISSUER = -9
    MANAGE_SELL_OFFER_BUY_NO_ISSUER = -10
    MANAGE_SELL_OFFER_NOT_FOUND = -11
    MANAGE_SELL_OFFER_LOW_RESERVE = -12

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ManageSellOfferResultCode:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ManageSellOfferResultCode:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ManageSellOfferResultCode:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



# Two ledgers average time + 1 sec
DEFAULT_GET_TIMEOUT_SECONDS = 11

# https://github.com/stellar/go/blob/7c4596db1ad107f770bd6e5a694ea5129440cb1e/services/horizon/internal/txsub/system.go#L314
_HORIZON_SUBMIT_TRANSACTION_API_TIMEOUT_SECONDS = 30
# POST is only used for submitting transactions to Horizon. Therefore we take the Horizon value with some grace.
DEFAULT_POST_TIMEOUT_SECONDS = _HORIZON_SUBMIT_TRANSACTION_API_TIMEOUT_SECONDS * 1.1



from decimal import Decimal
from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..asset import Asset
from ..muxed_account import MuxedAccount
from ..price import Price
from ..utils import raise_if_not_valid_amount
from .operation import Operation

__all__ = ["CreatePassiveSellOffer"]


class CreatePassiveSellOffer(Operation):
    """The :class:`CreatePassiveSellOffer` object, which represents a
    CreatePassiveSellOffer operation on Stellar's network.

    A passive sell offer is an offer that does not act on and take a reverse offer
    of equal price. Instead, they only take offers of lesser price. For
    example, if an offer exists to buy 5 BTC for 30 XLM, and you make a passive
    sell offer to buy 30 XLM for 5 BTC, your passive sell offer does not take the first
    offer.

    Note that regular offers made later than your passive sell offer can act on and
    take your passive sell offer, even if the regular offer is of the same price as
    your passive sell offer.

    Passive sell offers allow market makers to have zero spread. If you want to
    trade EUR for USD at 1:1 price and USD for EUR also at 1:1, you can create
    two passive sell offers so the two offers don't immediately act on each other.

    Once the passive sell offer is created, you can manage it like any other offer
    using the manage offer operation - see :class:`ManageBuyOffer` for more
    details.

    Threshold: Medium

    See `Create Passive Sell Offer <https://developers.stellar.org/docs/start/list-of-operations/#create-passive-sell-offer>`_ for more information.

    :param selling: What you're selling.
    :param buying: What you're buying.
    :param amount: The total amount you're selling.
    :param price: Price of 1 unit of `selling` in terms of `buying`.
    :param source: The source account for the operation. Defaults to the transaction's source account.

    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.CREATE_PASSIVE_SELL_OFFER
    )

    def __init__(
        self,
        selling: Asset,
        buying: Asset,
        amount: Union[str, Decimal],
        price: Union[Price, str, Decimal],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> None:
        super().__init__(source)
        if not isinstance(price, Price):
            price = Price.from_raw_price(price)

        self.selling: Asset = selling
        self.buying: Asset = buying
        self.amount: str = str(amount)
        self.price: Price = price
        raise_if_not_valid_amount(self.amount, "amount")

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        selling = self.selling.to_xdr_object()
        buying = self.buying.to_xdr_object()
        price = self.price.to_xdr_object()
        amount = stellar_xdr.Int64(Operation.to_xdr_amount(self.amount))
        create_passive_sell_offer_op = stellar_xdr.CreatePassiveSellOfferOp(
            selling, buying, amount, price
        )
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE,
            create_passive_sell_offer_op=create_passive_sell_offer_op,
        )
        return body

    @classmethod
    def from_xdr_object(
        cls, xdr_object: stellar_xdr.Operation
    ) -> "CreatePassiveSellOffer":
        """Creates a :class:`CreatePassiveSellOffer` object from an XDR Operation object."""
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.create_passive_sell_offer_op is not None
        selling = Asset.from_xdr_object(
            xdr_object.body.create_passive_sell_offer_op.selling
        )
        buying = Asset.from_xdr_object(
            xdr_object.body.create_passive_sell_offer_op.buying
        )
        amount = Operation.from_xdr_amount(
            xdr_object.body.create_passive_sell_offer_op.amount.int64
        )
        price = Price.from_xdr_object(
            xdr_object.body.create_passive_sell_offer_op.price
        )

        op = cls(
            source=source, selling=selling, buying=buying, amount=amount, price=price
        )
        return op

    def __repr__(self):
        return (
            f"<CreatePassiveSellOffer [selling={self.selling}, buying={self.buying}, "
            f"amount={self.amount}, price={self.price}, source={self.source}]>"
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["SetOptionsResultCode"]


class SetOptionsResultCode(IntEnum):
    """
    XDR Source Code::

        enum SetOptionsResultCode
        {
            // codes considered as "success" for the operation
            SET_OPTIONS_SUCCESS = 0,
            // codes considered as "failure" for the operation
            SET_OPTIONS_LOW_RESERVE = -1,      // not enough funds to add a signer
            SET_OPTIONS_TOO_MANY_SIGNERS = -2, // max number of signers already reached
            SET_OPTIONS_BAD_FLAGS = -3,        // invalid combination of clear/set flags
            SET_OPTIONS_INVALID_INFLATION = -4,      // inflation account does not exist
            SET_OPTIONS_CANT_CHANGE = -5,            // can no longer change this option
            SET_OPTIONS_UNKNOWN_FLAG = -6,           // can't set an unknown flag
            SET_OPTIONS_THRESHOLD_OUT_OF_RANGE = -7, // bad value for weight/threshold
            SET_OPTIONS_BAD_SIGNER = -8,             // signer cannot be masterkey
            SET_OPTIONS_INVALID_HOME_DOMAIN = -9,    // malformed home domain
            SET_OPTIONS_AUTH_REVOCABLE_REQUIRED =
                -10 // auth revocable is required for clawback
        };
    """

    SET_OPTIONS_SUCCESS = 0
    SET_OPTIONS_LOW_RESERVE = -1
    SET_OPTIONS_TOO_MANY_SIGNERS = -2
    SET_OPTIONS_BAD_FLAGS = -3
    SET_OPTIONS_INVALID_INFLATION = -4
    SET_OPTIONS_CANT_CHANGE = -5
    SET_OPTIONS_UNKNOWN_FLAG = -6
    SET_OPTIONS_THRESHOLD_OUT_OF_RANGE = -7
    SET_OPTIONS_BAD_SIGNER = -8
    SET_OPTIONS_INVALID_HOME_DOMAIN = -9
    SET_OPTIONS_AUTH_REVOCABLE_REQUIRED = -10

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SetOptionsResultCode:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SetOptionsResultCode:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SetOptionsResultCode:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



from abc import ABCMeta, abstractmethod
from typing import Any, AsyncGenerator, Dict

from .response import Response

__all__ = ["BaseAsyncClient"]


class BaseAsyncClient(metaclass=ABCMeta):
    """This is an abstract class,
    and if you want to implement your own asynchronous client, you **must** implement this class.

    """

    @abstractmethod
    async def get(self, url: str, params: Dict[str, str] = None) -> Response:
        """Perform HTTP GET request.

        :param url: the request url
        :param params: the request params
        :return: the response from server
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """
        pass  # pragma: no cover

    @abstractmethod
    async def post(
        self,
        url: str,
        data: Dict[str, str] = None,
        json_data: Dict[str, Any] = None,
    ) -> Response:
        """Perform HTTP POST request.

        :param url: the request url
        :param data: the data send to server
        :param json_data: the json data send to server
        :return: the response from server
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """
        pass  # pragma: no cover

    @abstractmethod
    def stream(
        self, url: str, params: Dict[str, str] = None
    ) -> AsyncGenerator[Dict[str, Any], None]:
        """Creates an EventSource that listens for incoming messages from the server.

        See `Horizon Response Format <https://developers.stellar.org/api/introduction/response-format/>`__

        See `MDN EventSource <https://developer.mozilla.org/en-US/docs/Web/API/EventSource>`_

        :param url: the request url
        :param params: the request params
        :return: a dict AsyncGenerator for server response
        :raise: :exc:`ConnectionError <stellar_sdk.exceptions.ConnectionError>`
        """
        pass  # pragma: no cover

    @abstractmethod
    async def close(self) -> None:
        pass  # pragma: no cover



from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..muxed_account import MuxedAccount
from .operation import Operation

__all__ = ["EndSponsoringFutureReserves"]


class EndSponsoringFutureReserves(Operation):
    """The :class:`EndSponsoringFutureReserves` object, which represents a EndSponsoringFutureReserves
    operation on Stellar's network.

    Terminates the current is-sponsoring-future-reserves-for relationship in which the source account is sponsored.
    See `Sponsored Reserves <https://developers.stellar.org/docs/glossary/sponsored-reserves/>`_ for more information.

    Threshold: Medium

    See `End Sponsoring Future Reserves <https://developers.stellar.org/docs/start/list-of-operations/#end-sponsoring-future-reserves>`_.

    :param source: The source account for the operation. Defaults to the transaction's source account.
    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.END_SPONSORING_FUTURE_RESERVES
    )

    def __init__(self, source: Optional[Union[MuxedAccount, str]] = None) -> None:
        super().__init__(source)

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        body = stellar_xdr.OperationBody(type=self._XDR_OPERATION_TYPE)
        return body

    @classmethod
    def from_xdr_object(
        cls, xdr_object: stellar_xdr.Operation
    ) -> "EndSponsoringFutureReserves":
        """Creates a :class:`EndSponsoringFutureReserves` object from an XDR Operation
        object.
        """
        source = Operation.get_source_from_xdr_obj(xdr_object)
        op = cls(source=source)
        return op

    def __repr__(self):
        return f"<EndSponsoringFutureReserves [source={self.source}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import Optional

from xdrlib3 import Packer, Unpacker

from .muxed_account import MuxedAccount
from .operation_body import OperationBody

__all__ = ["Operation"]


class Operation:
    """
    XDR Source Code::

        struct Operation
        {
            // sourceAccount is the account used to run the operation
            // if not set, the runtime defaults to "sourceAccount" specified at
            // the transaction level
            MuxedAccount* sourceAccount;

            union switch (OperationType type)
            {
            case CREATE_ACCOUNT:
                CreateAccountOp createAccountOp;
            case PAYMENT:
                PaymentOp paymentOp;
            case PATH_PAYMENT_STRICT_RECEIVE:
                PathPaymentStrictReceiveOp pathPaymentStrictReceiveOp;
            case MANAGE_SELL_OFFER:
                ManageSellOfferOp manageSellOfferOp;
            case CREATE_PASSIVE_SELL_OFFER:
                CreatePassiveSellOfferOp createPassiveSellOfferOp;
            case SET_OPTIONS:
                SetOptionsOp setOptionsOp;
            case CHANGE_TRUST:
                ChangeTrustOp changeTrustOp;
            case ALLOW_TRUST:
                AllowTrustOp allowTrustOp;
            case ACCOUNT_MERGE:
                MuxedAccount destination;
            case INFLATION:
                void;
            case MANAGE_DATA:
                ManageDataOp manageDataOp;
            case BUMP_SEQUENCE:
                BumpSequenceOp bumpSequenceOp;
            case MANAGE_BUY_OFFER:
                ManageBuyOfferOp manageBuyOfferOp;
            case PATH_PAYMENT_STRICT_SEND:
                PathPaymentStrictSendOp pathPaymentStrictSendOp;
            case CREATE_CLAIMABLE_BALANCE:
                CreateClaimableBalanceOp createClaimableBalanceOp;
            case CLAIM_CLAIMABLE_BALANCE:
                ClaimClaimableBalanceOp claimClaimableBalanceOp;
            case BEGIN_SPONSORING_FUTURE_RESERVES:
                BeginSponsoringFutureReservesOp beginSponsoringFutureReservesOp;
            case END_SPONSORING_FUTURE_RESERVES:
                void;
            case REVOKE_SPONSORSHIP:
                RevokeSponsorshipOp revokeSponsorshipOp;
            case CLAWBACK:
                ClawbackOp clawbackOp;
            case CLAWBACK_CLAIMABLE_BALANCE:
                ClawbackClaimableBalanceOp clawbackClaimableBalanceOp;
            case SET_TRUST_LINE_FLAGS:
                SetTrustLineFlagsOp setTrustLineFlagsOp;
            case LIQUIDITY_POOL_DEPOSIT:
                LiquidityPoolDepositOp liquidityPoolDepositOp;
            case LIQUIDITY_POOL_WITHDRAW:
                LiquidityPoolWithdrawOp liquidityPoolWithdrawOp;
            case INVOKE_HOST_FUNCTION:
                InvokeHostFunctionOp invokeHostFunctionOp;
            case EXTEND_FOOTPRINT_TTL:
                ExtendFootprintTTLOp extendFootprintTTLOp;
            case RESTORE_FOOTPRINT:
                RestoreFootprintOp restoreFootprintOp;
            }
            body;
        };
    """

    def __init__(
        self,
        source_account: Optional[MuxedAccount],
        body: OperationBody,
    ) -> None:
        self.source_account = source_account
        self.body = body

    def pack(self, packer: Packer) -> None:
        if self.source_account is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.source_account.pack(packer)
        self.body.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> Operation:
        source_account = (
            MuxedAccount.unpack(unpacker) if unpacker.unpack_uint() else None
        )
        body = OperationBody.unpack(unpacker)
        return cls(
            source_account=source_account,
            body=body,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> Operation:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> Operation:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.source_account,
                self.body,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.source_account == other.source_account and self.body == other.body

    def __repr__(self):
        out = [
            f"source_account={self.source_account}",
            f"body={self.body}",
        ]
        return f"<Operation [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import Optional

from xdrlib3 import Packer, Unpacker

from .contract_event_body import ContractEventBody
from .contract_event_type import ContractEventType
from .extension_point import ExtensionPoint
from .hash import Hash

__all__ = ["ContractEvent"]


class ContractEvent:
    """
    XDR Source Code::

        struct ContractEvent
        {
            // We can use this to add more fields, or because it
            // is first, to change ContractEvent into a union.
            ExtensionPoint ext;

            Hash* contractID;
            ContractEventType type;

            union switch (int v)
            {
            case 0:
                struct
                {
                    SCVal topics<>;
                    SCVal data;
                } v0;
            }
            body;
        };
    """

    def __init__(
        self,
        ext: ExtensionPoint,
        contract_id: Optional[Hash],
        type: ContractEventType,
        body: ContractEventBody,
    ) -> None:
        self.ext = ext
        self.contract_id = contract_id
        self.type = type
        self.body = body

    def pack(self, packer: Packer) -> None:
        self.ext.pack(packer)
        if self.contract_id is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.contract_id.pack(packer)
        self.type.pack(packer)
        self.body.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ContractEvent:
        ext = ExtensionPoint.unpack(unpacker)
        contract_id = Hash.unpack(unpacker) if unpacker.unpack_uint() else None
        type = ContractEventType.unpack(unpacker)
        body = ContractEventBody.unpack(unpacker)
        return cls(
            ext=ext,
            contract_id=contract_id,
            type=type,
            body=body,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ContractEvent:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ContractEvent:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ext,
                self.contract_id,
                self.type,
                self.body,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.ext == other.ext
            and self.contract_id == other.contract_id
            and self.type == other.type
            and self.body == other.body
        )

    def __repr__(self):
        out = [
            f"ext={self.ext}",
            f"contract_id={self.contract_id}",
            f"type={self.type}",
            f"body={self.body}",
        ]
        return f"<ContractEvent [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .account_merge_result import AccountMergeResult
from .allow_trust_result import AllowTrustResult
from .begin_sponsoring_future_reserves_result import BeginSponsoringFutureReservesResult
from .bump_sequence_result import BumpSequenceResult
from .change_trust_result import ChangeTrustResult
from .claim_claimable_balance_result import ClaimClaimableBalanceResult
from .clawback_claimable_balance_result import ClawbackClaimableBalanceResult
from .clawback_result import ClawbackResult
from .create_account_result import CreateAccountResult
from .create_claimable_balance_result import CreateClaimableBalanceResult
from .end_sponsoring_future_reserves_result import EndSponsoringFutureReservesResult
from .extend_footprint_ttl_result import ExtendFootprintTTLResult
from .inflation_result import InflationResult
from .invoke_host_function_result import InvokeHostFunctionResult
from .liquidity_pool_deposit_result import LiquidityPoolDepositResult
from .liquidity_pool_withdraw_result import LiquidityPoolWithdrawResult
from .manage_buy_offer_result import ManageBuyOfferResult
from .manage_data_result import ManageDataResult
from .manage_sell_offer_result import ManageSellOfferResult
from .operation_type import OperationType
from .path_payment_strict_receive_result import PathPaymentStrictReceiveResult
from .path_payment_strict_send_result import PathPaymentStrictSendResult
from .payment_result import PaymentResult
from .restore_footprint_result import RestoreFootprintResult
from .revoke_sponsorship_result import RevokeSponsorshipResult
from .set_options_result import SetOptionsResult
from .set_trust_line_flags_result import SetTrustLineFlagsResult

__all__ = ["OperationResultTr"]


class OperationResultTr:
    """
    XDR Source Code::

        union switch (OperationType type)
            {
            case CREATE_ACCOUNT:
                CreateAccountResult createAccountResult;
            case PAYMENT:
                PaymentResult paymentResult;
            case PATH_PAYMENT_STRICT_RECEIVE:
                PathPaymentStrictReceiveResult pathPaymentStrictReceiveResult;
            case MANAGE_SELL_OFFER:
                ManageSellOfferResult manageSellOfferResult;
            case CREATE_PASSIVE_SELL_OFFER:
                ManageSellOfferResult createPassiveSellOfferResult;
            case SET_OPTIONS:
                SetOptionsResult setOptionsResult;
            case CHANGE_TRUST:
                ChangeTrustResult changeTrustResult;
            case ALLOW_TRUST:
                AllowTrustResult allowTrustResult;
            case ACCOUNT_MERGE:
                AccountMergeResult accountMergeResult;
            case INFLATION:
                InflationResult inflationResult;
            case MANAGE_DATA:
                ManageDataResult manageDataResult;
            case BUMP_SEQUENCE:
                BumpSequenceResult bumpSeqResult;
            case MANAGE_BUY_OFFER:
                ManageBuyOfferResult manageBuyOfferResult;
            case PATH_PAYMENT_STRICT_SEND:
                PathPaymentStrictSendResult pathPaymentStrictSendResult;
            case CREATE_CLAIMABLE_BALANCE:
                CreateClaimableBalanceResult createClaimableBalanceResult;
            case CLAIM_CLAIMABLE_BALANCE:
                ClaimClaimableBalanceResult claimClaimableBalanceResult;
            case BEGIN_SPONSORING_FUTURE_RESERVES:
                BeginSponsoringFutureReservesResult beginSponsoringFutureReservesResult;
            case END_SPONSORING_FUTURE_RESERVES:
                EndSponsoringFutureReservesResult endSponsoringFutureReservesResult;
            case REVOKE_SPONSORSHIP:
                RevokeSponsorshipResult revokeSponsorshipResult;
            case CLAWBACK:
                ClawbackResult clawbackResult;
            case CLAWBACK_CLAIMABLE_BALANCE:
                ClawbackClaimableBalanceResult clawbackClaimableBalanceResult;
            case SET_TRUST_LINE_FLAGS:
                SetTrustLineFlagsResult setTrustLineFlagsResult;
            case LIQUIDITY_POOL_DEPOSIT:
                LiquidityPoolDepositResult liquidityPoolDepositResult;
            case LIQUIDITY_POOL_WITHDRAW:
                LiquidityPoolWithdrawResult liquidityPoolWithdrawResult;
            case INVOKE_HOST_FUNCTION:
                InvokeHostFunctionResult invokeHostFunctionResult;
            case EXTEND_FOOTPRINT_TTL:
                ExtendFootprintTTLResult extendFootprintTTLResult;
            case RESTORE_FOOTPRINT:
                RestoreFootprintResult restoreFootprintResult;
            }
    """

    def __init__(
        self,
        type: OperationType,
        create_account_result: CreateAccountResult = None,
        payment_result: PaymentResult = None,
        path_payment_strict_receive_result: PathPaymentStrictReceiveResult = None,
        manage_sell_offer_result: ManageSellOfferResult = None,
        create_passive_sell_offer_result: ManageSellOfferResult = None,
        set_options_result: SetOptionsResult = None,
        change_trust_result: ChangeTrustResult = None,
        allow_trust_result: AllowTrustResult = None,
        account_merge_result: AccountMergeResult = None,
        inflation_result: InflationResult = None,
        manage_data_result: ManageDataResult = None,
        bump_seq_result: BumpSequenceResult = None,
        manage_buy_offer_result: ManageBuyOfferResult = None,
        path_payment_strict_send_result: PathPaymentStrictSendResult = None,
        create_claimable_balance_result: CreateClaimableBalanceResult = None,
        claim_claimable_balance_result: ClaimClaimableBalanceResult = None,
        begin_sponsoring_future_reserves_result: BeginSponsoringFutureReservesResult = None,
        end_sponsoring_future_reserves_result: EndSponsoringFutureReservesResult = None,
        revoke_sponsorship_result: RevokeSponsorshipResult = None,
        clawback_result: ClawbackResult = None,
        clawback_claimable_balance_result: ClawbackClaimableBalanceResult = None,
        set_trust_line_flags_result: SetTrustLineFlagsResult = None,
        liquidity_pool_deposit_result: LiquidityPoolDepositResult = None,
        liquidity_pool_withdraw_result: LiquidityPoolWithdrawResult = None,
        invoke_host_function_result: InvokeHostFunctionResult = None,
        extend_footprint_ttl_result: ExtendFootprintTTLResult = None,
        restore_footprint_result: RestoreFootprintResult = None,
    ) -> None:
        self.type = type
        self.create_account_result = create_account_result
        self.payment_result = payment_result
        self.path_payment_strict_receive_result = path_payment_strict_receive_result
        self.manage_sell_offer_result = manage_sell_offer_result
        self.create_passive_sell_offer_result = create_passive_sell_offer_result
        self.set_options_result = set_options_result
        self.change_trust_result = change_trust_result
        self.allow_trust_result = allow_trust_result
        self.account_merge_result = account_merge_result
        self.inflation_result = inflation_result
        self.manage_data_result = manage_data_result
        self.bump_seq_result = bump_seq_result
        self.manage_buy_offer_result = manage_buy_offer_result
        self.path_payment_strict_send_result = path_payment_strict_send_result
        self.create_claimable_balance_result = create_claimable_balance_result
        self.claim_claimable_balance_result = claim_claimable_balance_result
        self.begin_sponsoring_future_reserves_result = (
            begin_sponsoring_future_reserves_result
        )
        self.end_sponsoring_future_reserves_result = (
            end_sponsoring_future_reserves_result
        )
        self.revoke_sponsorship_result = revoke_sponsorship_result
        self.clawback_result = clawback_result
        self.clawback_claimable_balance_result = clawback_claimable_balance_result
        self.set_trust_line_flags_result = set_trust_line_flags_result
        self.liquidity_pool_deposit_result = liquidity_pool_deposit_result
        self.liquidity_pool_withdraw_result = liquidity_pool_withdraw_result
        self.invoke_host_function_result = invoke_host_function_result
        self.extend_footprint_ttl_result = extend_footprint_ttl_result
        self.restore_footprint_result = restore_footprint_result

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == OperationType.CREATE_ACCOUNT:
            if self.create_account_result is None:
                raise ValueError("create_account_result should not be None.")
            self.create_account_result.pack(packer)
            return
        if self.type == OperationType.PAYMENT:
            if self.payment_result is None:
                raise ValueError("payment_result should not be None.")
            self.payment_result.pack(packer)
            return
        if self.type == OperationType.PATH_PAYMENT_STRICT_RECEIVE:
            if self.path_payment_strict_receive_result is None:
                raise ValueError(
                    "path_payment_strict_receive_result should not be None."
                )
            self.path_payment_strict_receive_result.pack(packer)
            return
        if self.type == OperationType.MANAGE_SELL_OFFER:
            if self.manage_sell_offer_result is None:
                raise ValueError("manage_sell_offer_result should not be None.")
            self.manage_sell_offer_result.pack(packer)
            return
        if self.type == OperationType.CREATE_PASSIVE_SELL_OFFER:
            if self.create_passive_sell_offer_result is None:
                raise ValueError("create_passive_sell_offer_result should not be None.")
            self.create_passive_sell_offer_result.pack(packer)
            return
        if self.type == OperationType.SET_OPTIONS:
            if self.set_options_result is None:
                raise ValueError("set_options_result should not be None.")
            self.set_options_result.pack(packer)
            return
        if self.type == OperationType.CHANGE_TRUST:
            if self.change_trust_result is None:
                raise ValueError("change_trust_result should not be None.")
            self.change_trust_result.pack(packer)
            return
        if self.type == OperationType.ALLOW_TRUST:
            if self.allow_trust_result is None:
                raise ValueError("allow_trust_result should not be None.")
            self.allow_trust_result.pack(packer)
            return
        if self.type == OperationType.ACCOUNT_MERGE:
            if self.account_merge_result is None:
                raise ValueError("account_merge_result should not be None.")
            self.account_merge_result.pack(packer)
            return
        if self.type == OperationType.INFLATION:
            if self.inflation_result is None:
                raise ValueError("inflation_result should not be None.")
            self.inflation_result.pack(packer)
            return
        if self.type == OperationType.MANAGE_DATA:
            if self.manage_data_result is None:
                raise ValueError("manage_data_result should not be None.")
            self.manage_data_result.pack(packer)
            return
        if self.type == OperationType.BUMP_SEQUENCE:
            if self.bump_seq_result is None:
                raise ValueError("bump_seq_result should not be None.")
            self.bump_seq_result.pack(packer)
            return
        if self.type == OperationType.MANAGE_BUY_OFFER:
            if self.manage_buy_offer_result is None:
                raise ValueError("manage_buy_offer_result should not be None.")
            self.manage_buy_offer_result.pack(packer)
            return
        if self.type == OperationType.PATH_PAYMENT_STRICT_SEND:
            if self.path_payment_strict_send_result is None:
                raise ValueError("path_payment_strict_send_result should not be None.")
            self.path_payment_strict_send_result.pack(packer)
            return
        if self.type == OperationType.CREATE_CLAIMABLE_BALANCE:
            if self.create_claimable_balance_result is None:
                raise ValueError("create_claimable_balance_result should not be None.")
            self.create_claimable_balance_result.pack(packer)
            return
        if self.type == OperationType.CLAIM_CLAIMABLE_BALANCE:
            if self.claim_claimable_balance_result is None:
                raise ValueError("claim_claimable_balance_result should not be None.")
            self.claim_claimable_balance_result.pack(packer)
            return
        if self.type == OperationType.BEGIN_SPONSORING_FUTURE_RESERVES:
            if self.begin_sponsoring_future_reserves_result is None:
                raise ValueError(
                    "begin_sponsoring_future_reserves_result should not be None."
                )
            self.begin_sponsoring_future_reserves_result.pack(packer)
            return
        if self.type == OperationType.END_SPONSORING_FUTURE_RESERVES:
            if self.end_sponsoring_future_reserves_result is None:
                raise ValueError(
                    "end_sponsoring_future_reserves_result should not be None."
                )
            self.end_sponsoring_future_reserves_result.pack(packer)
            return
        if self.type == OperationType.REVOKE_SPONSORSHIP:
            if self.revoke_sponsorship_result is None:
                raise ValueError("revoke_sponsorship_result should not be None.")
            self.revoke_sponsorship_result.pack(packer)
            return
        if self.type == OperationType.CLAWBACK:
            if self.clawback_result is None:
                raise ValueError("clawback_result should not be None.")
            self.clawback_result.pack(packer)
            return
        if self.type == OperationType.CLAWBACK_CLAIMABLE_BALANCE:
            if self.clawback_claimable_balance_result is None:
                raise ValueError(
                    "clawback_claimable_balance_result should not be None."
                )
            self.clawback_claimable_balance_result.pack(packer)
            return
        if self.type == OperationType.SET_TRUST_LINE_FLAGS:
            if self.set_trust_line_flags_result is None:
                raise ValueError("set_trust_line_flags_result should not be None.")
            self.set_trust_line_flags_result.pack(packer)
            return
        if self.type == OperationType.LIQUIDITY_POOL_DEPOSIT:
            if self.liquidity_pool_deposit_result is None:
                raise ValueError("liquidity_pool_deposit_result should not be None.")
            self.liquidity_pool_deposit_result.pack(packer)
            return
        if self.type == OperationType.LIQUIDITY_POOL_WITHDRAW:
            if self.liquidity_pool_withdraw_result is None:
                raise ValueError("liquidity_pool_withdraw_result should not be None.")
            self.liquidity_pool_withdraw_result.pack(packer)
            return
        if self.type == OperationType.INVOKE_HOST_FUNCTION:
            if self.invoke_host_function_result is None:
                raise ValueError("invoke_host_function_result should not be None.")
            self.invoke_host_function_result.pack(packer)
            return
        if self.type == OperationType.EXTEND_FOOTPRINT_TTL:
            if self.extend_footprint_ttl_result is None:
                raise ValueError("extend_footprint_ttl_result should not be None.")
            self.extend_footprint_ttl_result.pack(packer)
            return
        if self.type == OperationType.RESTORE_FOOTPRINT:
            if self.restore_footprint_result is None:
                raise ValueError("restore_footprint_result should not be None.")
            self.restore_footprint_result.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> OperationResultTr:
        type = OperationType.unpack(unpacker)
        if type == OperationType.CREATE_ACCOUNT:
            create_account_result = CreateAccountResult.unpack(unpacker)
            return cls(type=type, create_account_result=create_account_result)
        if type == OperationType.PAYMENT:
            payment_result = PaymentResult.unpack(unpacker)
            return cls(type=type, payment_result=payment_result)
        if type == OperationType.PATH_PAYMENT_STRICT_RECEIVE:
            path_payment_strict_receive_result = PathPaymentStrictReceiveResult.unpack(
                unpacker
            )
            return cls(
                type=type,
                path_payment_strict_receive_result=path_payment_strict_receive_result,
            )
        if type == OperationType.MANAGE_SELL_OFFER:
            manage_sell_offer_result = ManageSellOfferResult.unpack(unpacker)
            return cls(type=type, manage_sell_offer_result=manage_sell_offer_result)
        if type == OperationType.CREATE_PASSIVE_SELL_OFFER:
            create_passive_sell_offer_result = ManageSellOfferResult.unpack(unpacker)
            return cls(
                type=type,
                create_passive_sell_offer_result=create_passive_sell_offer_result,
            )
        if type == OperationType.SET_OPTIONS:
            set_options_result = SetOptionsResult.unpack(unpacker)
            return cls(type=type, set_options_result=set_options_result)
        if type == OperationType.CHANGE_TRUST:
            change_trust_result = ChangeTrustResult.unpack(unpacker)
            return cls(type=type, change_trust_result=change_trust_result)
        if type == OperationType.ALLOW_TRUST:
            allow_trust_result = AllowTrustResult.unpack(unpacker)
            return cls(type=type, allow_trust_result=allow_trust_result)
        if type == OperationType.ACCOUNT_MERGE:
            account_merge_result = AccountMergeResult.unpack(unpacker)
            return cls(type=type, account_merge_result=account_merge_result)
        if type == OperationType.INFLATION:
            inflation_result = InflationResult.unpack(unpacker)
            return cls(type=type, inflation_result=inflation_result)
        if type == OperationType.MANAGE_DATA:
            manage_data_result = ManageDataResult.unpack(unpacker)
            return cls(type=type, manage_data_result=manage_data_result)
        if type == OperationType.BUMP_SEQUENCE:
            bump_seq_result = BumpSequenceResult.unpack(unpacker)
            return cls(type=type, bump_seq_result=bump_seq_result)
        if type == OperationType.MANAGE_BUY_OFFER:
            manage_buy_offer_result = ManageBuyOfferResult.unpack(unpacker)
            return cls(type=type, manage_buy_offer_result=manage_buy_offer_result)
        if type == OperationType.PATH_PAYMENT_STRICT_SEND:
            path_payment_strict_send_result = PathPaymentStrictSendResult.unpack(
                unpacker
            )
            return cls(
                type=type,
                path_payment_strict_send_result=path_payment_strict_send_result,
            )
        if type == OperationType.CREATE_CLAIMABLE_BALANCE:
            create_claimable_balance_result = CreateClaimableBalanceResult.unpack(
                unpacker
            )
            return cls(
                type=type,
                create_claimable_balance_result=create_claimable_balance_result,
            )
        if type == OperationType.CLAIM_CLAIMABLE_BALANCE:
            claim_claimable_balance_result = ClaimClaimableBalanceResult.unpack(
                unpacker
            )
            return cls(
                type=type, claim_claimable_balance_result=claim_claimable_balance_result
            )
        if type == OperationType.BEGIN_SPONSORING_FUTURE_RESERVES:
            begin_sponsoring_future_reserves_result = (
                BeginSponsoringFutureReservesResult.unpack(unpacker)
            )
            return cls(
                type=type,
                begin_sponsoring_future_reserves_result=begin_sponsoring_future_reserves_result,
            )
        if type == OperationType.END_SPONSORING_FUTURE_RESERVES:
            end_sponsoring_future_reserves_result = (
                EndSponsoringFutureReservesResult.unpack(unpacker)
            )
            return cls(
                type=type,
                end_sponsoring_future_reserves_result=end_sponsoring_future_reserves_result,
            )
        if type == OperationType.REVOKE_SPONSORSHIP:
            revoke_sponsorship_result = RevokeSponsorshipResult.unpack(unpacker)
            return cls(type=type, revoke_sponsorship_result=revoke_sponsorship_result)
        if type == OperationType.CLAWBACK:
            clawback_result = ClawbackResult.unpack(unpacker)
            return cls(type=type, clawback_result=clawback_result)
        if type == OperationType.CLAWBACK_CLAIMABLE_BALANCE:
            clawback_claimable_balance_result = ClawbackClaimableBalanceResult.unpack(
                unpacker
            )
            return cls(
                type=type,
                clawback_claimable_balance_result=clawback_claimable_balance_result,
            )
        if type == OperationType.SET_TRUST_LINE_FLAGS:
            set_trust_line_flags_result = SetTrustLineFlagsResult.unpack(unpacker)
            return cls(
                type=type, set_trust_line_flags_result=set_trust_line_flags_result
            )
        if type == OperationType.LIQUIDITY_POOL_DEPOSIT:
            liquidity_pool_deposit_result = LiquidityPoolDepositResult.unpack(unpacker)
            return cls(
                type=type, liquidity_pool_deposit_result=liquidity_pool_deposit_result
            )
        if type == OperationType.LIQUIDITY_POOL_WITHDRAW:
            liquidity_pool_withdraw_result = LiquidityPoolWithdrawResult.unpack(
                unpacker
            )
            return cls(
                type=type, liquidity_pool_withdraw_result=liquidity_pool_withdraw_result
            )
        if type == OperationType.INVOKE_HOST_FUNCTION:
            invoke_host_function_result = InvokeHostFunctionResult.unpack(unpacker)
            return cls(
                type=type, invoke_host_function_result=invoke_host_function_result
            )
        if type == OperationType.EXTEND_FOOTPRINT_TTL:
            extend_footprint_ttl_result = ExtendFootprintTTLResult.unpack(unpacker)
            return cls(
                type=type, extend_footprint_ttl_result=extend_footprint_ttl_result
            )
        if type == OperationType.RESTORE_FOOTPRINT:
            restore_footprint_result = RestoreFootprintResult.unpack(unpacker)
            return cls(type=type, restore_footprint_result=restore_footprint_result)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> OperationResultTr:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> OperationResultTr:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.create_account_result,
                self.payment_result,
                self.path_payment_strict_receive_result,
                self.manage_sell_offer_result,
                self.create_passive_sell_offer_result,
                self.set_options_result,
                self.change_trust_result,
                self.allow_trust_result,
                self.account_merge_result,
                self.inflation_result,
                self.manage_data_result,
                self.bump_seq_result,
                self.manage_buy_offer_result,
                self.path_payment_strict_send_result,
                self.create_claimable_balance_result,
                self.claim_claimable_balance_result,
                self.begin_sponsoring_future_reserves_result,
                self.end_sponsoring_future_reserves_result,
                self.revoke_sponsorship_result,
                self.clawback_result,
                self.clawback_claimable_balance_result,
                self.set_trust_line_flags_result,
                self.liquidity_pool_deposit_result,
                self.liquidity_pool_withdraw_result,
                self.invoke_host_function_result,
                self.extend_footprint_ttl_result,
                self.restore_footprint_result,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.create_account_result == other.create_account_result
            and self.payment_result == other.payment_result
            and self.path_payment_strict_receive_result
            == other.path_payment_strict_receive_result
            and self.manage_sell_offer_result == other.manage_sell_offer_result
            and self.create_passive_sell_offer_result
            == other.create_passive_sell_offer_result
            and self.set_options_result == other.set_options_result
            and self.change_trust_result == other.change_trust_result
            and self.allow_trust_result == other.allow_trust_result
            and self.account_merge_result == other.account_merge_result
            and self.inflation_result == other.inflation_result
            and self.manage_data_result == other.manage_data_result
            and self.bump_seq_result == other.bump_seq_result
            and self.manage_buy_offer_result == other.manage_buy_offer_result
            and self.path_payment_strict_send_result
            == other.path_payment_strict_send_result
            and self.create_claimable_balance_result
            == other.create_claimable_balance_result
            and self.claim_claimable_balance_result
            == other.claim_claimable_balance_result
            and self.begin_sponsoring_future_reserves_result
            == other.begin_sponsoring_future_reserves_result
            and self.end_sponsoring_future_reserves_result
            == other.end_sponsoring_future_reserves_result
            and self.revoke_sponsorship_result == other.revoke_sponsorship_result
            and self.clawback_result == other.clawback_result
            and self.clawback_claimable_balance_result
            == other.clawback_claimable_balance_result
            and self.set_trust_line_flags_result == other.set_trust_line_flags_result
            and self.liquidity_pool_deposit_result
            == other.liquidity_pool_deposit_result
            and self.liquidity_pool_withdraw_result
            == other.liquidity_pool_withdraw_result
            and self.invoke_host_function_result == other.invoke_host_function_result
            and self.extend_footprint_ttl_result == other.extend_footprint_ttl_result
            and self.restore_footprint_result == other.restore_footprint_result
        )

    def __repr__(self):
        out = []
        out.append(f"type={self.type}")
        (
            out.append(f"create_account_result={self.create_account_result}")
            if self.create_account_result is not None
            else None
        )
        (
            out.append(f"payment_result={self.payment_result}")
            if self.payment_result is not None
            else None
        )
        (
            out.append(
                f"path_payment_strict_receive_result={self.path_payment_strict_receive_result}"
            )
            if self.path_payment_strict_receive_result is not None
            else None
        )
        (
            out.append(f"manage_sell_offer_result={self.manage_sell_offer_result}")
            if self.manage_sell_offer_result is not None
            else None
        )
        (
            out.append(
                f"create_passive_sell_offer_result={self.create_passive_sell_offer_result}"
            )
            if self.create_passive_sell_offer_result is not None
            else None
        )
        (
            out.append(f"set_options_result={self.set_options_result}")
            if self.set_options_result is not None
            else None
        )
        (
            out.append(f"change_trust_result={self.change_trust_result}")
            if self.change_trust_result is not None
            else None
        )
        (
            out.append(f"allow_trust_result={self.allow_trust_result}")
            if self.allow_trust_result is not None
            else None
        )
        (
            out.append(f"account_merge_result={self.account_merge_result}")
            if self.account_merge_result is not None
            else None
        )
        (
            out.append(f"inflation_result={self.inflation_result}")
            if self.inflation_result is not None
            else None
        )
        (
            out.append(f"manage_data_result={self.manage_data_result}")
            if self.manage_data_result is not None
            else None
        )
        (
            out.append(f"bump_seq_result={self.bump_seq_result}")
            if self.bump_seq_result is not None
            else None
        )
        (
            out.append(f"manage_buy_offer_result={self.manage_buy_offer_result}")
            if self.manage_buy_offer_result is not None
            else None
        )
        (
            out.append(
                f"path_payment_strict_send_result={self.path_payment_strict_send_result}"
            )
            if self.path_payment_strict_send_result is not None
            else None
        )
        (
            out.append(
                f"create_claimable_balance_result={self.create_claimable_balance_result}"
            )
            if self.create_claimable_balance_result is not None
            else None
        )
        (
            out.append(
                f"claim_claimable_balance_result={self.claim_claimable_balance_result}"
            )
            if self.claim_claimable_balance_result is not None
            else None
        )
        (
            out.append(
                f"begin_sponsoring_future_reserves_result={self.begin_sponsoring_future_reserves_result}"
            )
            if self.begin_sponsoring_future_reserves_result is not None
            else None
        )
        (
            out.append(
                f"end_sponsoring_future_reserves_result={self.end_sponsoring_future_reserves_result}"
            )
            if self.end_sponsoring_future_reserves_result is not None
            else None
        )
        (
            out.append(f"revoke_sponsorship_result={self.revoke_sponsorship_result}")
            if self.revoke_sponsorship_result is not None
            else None
        )
        (
            out.append(f"clawback_result={self.clawback_result}")
            if self.clawback_result is not None
            else None
        )
        (
            out.append(
                f"clawback_claimable_balance_result={self.clawback_claimable_balance_result}"
            )
            if self.clawback_claimable_balance_result is not None
            else None
        )
        (
            out.append(
                f"set_trust_line_flags_result={self.set_trust_line_flags_result}"
            )
            if self.set_trust_line_flags_result is not None
            else None
        )
        (
            out.append(
                f"liquidity_pool_deposit_result={self.liquidity_pool_deposit_result}"
            )
            if self.liquidity_pool_deposit_result is not None
            else None
        )
        (
            out.append(
                f"liquidity_pool_withdraw_result={self.liquidity_pool_withdraw_result}"
            )
            if self.liquidity_pool_withdraw_result is not None
            else None
        )
        (
            out.append(
                f"invoke_host_function_result={self.invoke_host_function_result}"
            )
            if self.invoke_host_function_result is not None
            else None
        )
        (
            out.append(
                f"extend_footprint_ttl_result={self.extend_footprint_ttl_result}"
            )
            if self.extend_footprint_ttl_result is not None
            else None
        )
        (
            out.append(f"restore_footprint_result={self.restore_footprint_result}")
            if self.restore_footprint_result is not None
            else None
        )
        return f"<OperationResultTr [{', '.join(out)}]>"



from ...asset import Asset
from ...call_builder.base import BaseTradeAggregationsCallBuilder
from ...call_builder.call_builder_sync.base_call_builder import BaseCallBuilder
from ...client.base_sync_client import BaseSyncClient

__all__ = ["TradeAggregationsCallBuilder"]


class TradeAggregationsCallBuilder(BaseCallBuilder, BaseTradeAggregationsCallBuilder):
    """Creates a new :class:`TradeAggregationsCallBuilder` pointed to server defined by horizon_url.
    Do not create this object directly, use :func:`stellar_sdk.Server.trade_aggregations`.

    Trade Aggregations facilitate efficient gathering of historical trade data.

    See `List Trade Aggregations <https://developers.stellar.org/api/aggregations/trade-aggregations/list/>`__ for more information.

    :param horizon_url: Horizon server URL.
    :param client: The client instance used to send request.
    :param base: base asset
    :param counter: counter asset
    :param resolution: segment duration as millis since epoch. *Supported values
        are 1 minute (60000), 5 minutes (300000), 15 minutes (900000),
        1 hour (3600000), 1 day (86400000) and 1 week (604800000).*
    :param start_time: lower time boundary represented as millis since epoch
    :param end_time: upper time boundary represented as millis since epoch
    :param offset: segments can be offset using this parameter.
        Expressed in milliseconds. *Can only be used if the resolution is greater than 1 hour.
        Value must be in whole hours, less than the provided resolution, and less than 24 hours.*
    """

    def __init__(
        self,
        horizon_url: str,
        client: BaseSyncClient,
        base: Asset,
        counter: Asset,
        resolution: int,
        start_time: int = None,
        end_time: int = None,
        offset: int = None,
    ) -> None:
        super().__init__(  # type: ignore[call-arg]
            horizon_url=horizon_url,
            client=client,
            base=base,
            counter=counter,
            resolution=resolution,
            start_time=start_time,
            end_time=end_time,
            offset=offset,
        )

    def stream(
        self,
    ):
        """This endpoint does not support streaming."""
        raise NotImplementedError("Streaming is not supported.")



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .ledger_key import LedgerKey

__all__ = ["LedgerFootprint"]


class LedgerFootprint:
    """
    XDR Source Code::

        struct LedgerFootprint
        {
            LedgerKey readOnly<>;
            LedgerKey readWrite<>;
        };
    """

    def __init__(
        self,
        read_only: List[LedgerKey],
        read_write: List[LedgerKey],
    ) -> None:
        _expect_max_length = 4294967295
        if read_only and len(read_only) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `read_only` should be {_expect_max_length}, but got {len(read_only)}."
            )
        _expect_max_length = 4294967295
        if read_write and len(read_write) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `read_write` should be {_expect_max_length}, but got {len(read_write)}."
            )
        self.read_only = read_only
        self.read_write = read_write

    def pack(self, packer: Packer) -> None:
        packer.pack_uint(len(self.read_only))
        for read_only_item in self.read_only:
            read_only_item.pack(packer)
        packer.pack_uint(len(self.read_write))
        for read_write_item in self.read_write:
            read_write_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LedgerFootprint:
        length = unpacker.unpack_uint()
        read_only = []
        for _ in range(length):
            read_only.append(LedgerKey.unpack(unpacker))
        length = unpacker.unpack_uint()
        read_write = []
        for _ in range(length):
            read_write.append(LedgerKey.unpack(unpacker))
        return cls(
            read_only=read_only,
            read_write=read_write,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LedgerFootprint:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LedgerFootprint:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.read_only,
                self.read_write,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.read_only == other.read_only and self.read_write == other.read_write

    def __repr__(self):
        out = [
            f"read_only={self.read_only}",
            f"read_write={self.read_write}",
        ]
        return f"<LedgerFootprint [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["SCValType"]


class SCValType(IntEnum):
    """
    XDR Source Code::

        enum SCValType
        {
            SCV_BOOL = 0,
            SCV_VOID = 1,
            SCV_ERROR = 2,

            // 32 bits is the smallest type in WASM or XDR; no need for u8/u16.
            SCV_U32 = 3,
            SCV_I32 = 4,

            // 64 bits is naturally supported by both WASM and XDR also.
            SCV_U64 = 5,
            SCV_I64 = 6,

            // Time-related u64 subtypes with their own functions and formatting.
            SCV_TIMEPOINT = 7,
            SCV_DURATION = 8,

            // 128 bits is naturally supported by Rust and we use it for Soroban
            // fixed-point arithmetic prices / balances / similar "quantities". These
            // are represented in XDR as a pair of 2 u64s.
            SCV_U128 = 9,
            SCV_I128 = 10,

            // 256 bits is the size of sha256 output, ed25519 keys, and the EVM machine
            // word, so for interop use we include this even though it requires a small
            // amount of Rust guest and/or host library code.
            SCV_U256 = 11,
            SCV_I256 = 12,

            // Bytes come in 3 flavors, 2 of which have meaningfully different
            // formatting and validity-checking / domain-restriction.
            SCV_BYTES = 13,
            SCV_STRING = 14,
            SCV_SYMBOL = 15,

            // Vecs and maps are just polymorphic containers of other ScVals.
            SCV_VEC = 16,
            SCV_MAP = 17,

            // Address is the universal identifier for contracts and classic
            // accounts.
            SCV_ADDRESS = 18,

            // The following are the internal SCVal variants that are not
            // exposed to the contracts.
            SCV_CONTRACT_INSTANCE = 19,

            // SCV_LEDGER_KEY_CONTRACT_INSTANCE and SCV_LEDGER_KEY_NONCE are unique
            // symbolic SCVals used as the key for ledger entries for a contract's
            // instance and an address' nonce, respectively.
            SCV_LEDGER_KEY_CONTRACT_INSTANCE = 20,
            SCV_LEDGER_KEY_NONCE = 21
        };
    """

    SCV_BOOL = 0
    SCV_VOID = 1
    SCV_ERROR = 2
    SCV_U32 = 3
    SCV_I32 = 4
    SCV_U64 = 5
    SCV_I64 = 6
    SCV_TIMEPOINT = 7
    SCV_DURATION = 8
    SCV_U128 = 9
    SCV_I128 = 10
    SCV_U256 = 11
    SCV_I256 = 12
    SCV_BYTES = 13
    SCV_STRING = 14
    SCV_SYMBOL = 15
    SCV_VEC = 16
    SCV_MAP = 17
    SCV_ADDRESS = 18
    SCV_CONTRACT_INSTANCE = 19
    SCV_LEDGER_KEY_CONTRACT_INSTANCE = 20
    SCV_LEDGER_KEY_NONCE = 21

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCValType:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCValType:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCValType:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .asset import Asset
from .path_payment_strict_receive_result_code import PathPaymentStrictReceiveResultCode
from .path_payment_strict_receive_result_success import (
    PathPaymentStrictReceiveResultSuccess,
)

__all__ = ["PathPaymentStrictReceiveResult"]


class PathPaymentStrictReceiveResult:
    """
    XDR Source Code::

        union PathPaymentStrictReceiveResult switch (
            PathPaymentStrictReceiveResultCode code)
        {
        case PATH_PAYMENT_STRICT_RECEIVE_SUCCESS:
            struct
            {
                ClaimAtom offers<>;
                SimplePaymentResult last;
            } success;
        case PATH_PAYMENT_STRICT_RECEIVE_MALFORMED:
        case PATH_PAYMENT_STRICT_RECEIVE_UNDERFUNDED:
        case PATH_PAYMENT_STRICT_RECEIVE_SRC_NO_TRUST:
        case PATH_PAYMENT_STRICT_RECEIVE_SRC_NOT_AUTHORIZED:
        case PATH_PAYMENT_STRICT_RECEIVE_NO_DESTINATION:
        case PATH_PAYMENT_STRICT_RECEIVE_NO_TRUST:
        case PATH_PAYMENT_STRICT_RECEIVE_NOT_AUTHORIZED:
        case PATH_PAYMENT_STRICT_RECEIVE_LINE_FULL:
            void;
        case PATH_PAYMENT_STRICT_RECEIVE_NO_ISSUER:
            Asset noIssuer; // the asset that caused the error
        case PATH_PAYMENT_STRICT_RECEIVE_TOO_FEW_OFFERS:
        case PATH_PAYMENT_STRICT_RECEIVE_OFFER_CROSS_SELF:
        case PATH_PAYMENT_STRICT_RECEIVE_OVER_SENDMAX:
            void;
        };
    """

    def __init__(
        self,
        code: PathPaymentStrictReceiveResultCode,
        success: PathPaymentStrictReceiveResultSuccess = None,
        no_issuer: Asset = None,
    ) -> None:
        self.code = code
        self.success = success
        self.no_issuer = no_issuer

    def pack(self, packer: Packer) -> None:
        self.code.pack(packer)
        if (
            self.code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_SUCCESS
        ):
            if self.success is None:
                raise ValueError("success should not be None.")
            self.success.pack(packer)
            return
        if (
            self.code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_MALFORMED
        ):
            return
        if (
            self.code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_UNDERFUNDED
        ):
            return
        if (
            self.code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_SRC_NO_TRUST
        ):
            return
        if (
            self.code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_SRC_NOT_AUTHORIZED
        ):
            return
        if (
            self.code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_NO_DESTINATION
        ):
            return
        if (
            self.code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_NO_TRUST
        ):
            return
        if (
            self.code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_NOT_AUTHORIZED
        ):
            return
        if (
            self.code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_LINE_FULL
        ):
            return
        if (
            self.code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_NO_ISSUER
        ):
            if self.no_issuer is None:
                raise ValueError("no_issuer should not be None.")
            self.no_issuer.pack(packer)
            return
        if (
            self.code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_TOO_FEW_OFFERS
        ):
            return
        if (
            self.code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_OFFER_CROSS_SELF
        ):
            return
        if (
            self.code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_OVER_SENDMAX
        ):
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> PathPaymentStrictReceiveResult:
        code = PathPaymentStrictReceiveResultCode.unpack(unpacker)
        if (
            code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_SUCCESS
        ):
            success = PathPaymentStrictReceiveResultSuccess.unpack(unpacker)
            return cls(code=code, success=success)
        if (
            code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_MALFORMED
        ):
            return cls(code=code)
        if (
            code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_UNDERFUNDED
        ):
            return cls(code=code)
        if (
            code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_SRC_NO_TRUST
        ):
            return cls(code=code)
        if (
            code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_SRC_NOT_AUTHORIZED
        ):
            return cls(code=code)
        if (
            code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_NO_DESTINATION
        ):
            return cls(code=code)
        if (
            code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_NO_TRUST
        ):
            return cls(code=code)
        if (
            code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_NOT_AUTHORIZED
        ):
            return cls(code=code)
        if (
            code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_LINE_FULL
        ):
            return cls(code=code)
        if (
            code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_NO_ISSUER
        ):
            no_issuer = Asset.unpack(unpacker)
            return cls(code=code, no_issuer=no_issuer)
        if (
            code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_TOO_FEW_OFFERS
        ):
            return cls(code=code)
        if (
            code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_OFFER_CROSS_SELF
        ):
            return cls(code=code)
        if (
            code
            == PathPaymentStrictReceiveResultCode.PATH_PAYMENT_STRICT_RECEIVE_OVER_SENDMAX
        ):
            return cls(code=code)
        return cls(code=code)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> PathPaymentStrictReceiveResult:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> PathPaymentStrictReceiveResult:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.code,
                self.success,
                self.no_issuer,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.code == other.code
            and self.success == other.success
            and self.no_issuer == other.no_issuer
        )

    def __repr__(self):
        out = []
        out.append(f"code={self.code}")
        out.append(f"success={self.success}") if self.success is not None else None
        (
            out.append(f"no_issuer={self.no_issuer}")
            if self.no_issuer is not None
            else None
        )
        return f"<PathPaymentStrictReceiveResult [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .asset import Asset
from .int64 import Int64
from .muxed_account import MuxedAccount

__all__ = ["ClawbackOp"]


class ClawbackOp:
    """
    XDR Source Code::

        struct ClawbackOp
        {
            Asset asset;
            MuxedAccount from_;
            int64 amount;
        };
    """

    def __init__(
        self,
        asset: Asset,
        from_: MuxedAccount,
        amount: Int64,
    ) -> None:
        self.asset = asset
        self.from_ = from_
        self.amount = amount

    def pack(self, packer: Packer) -> None:
        self.asset.pack(packer)
        self.from_.pack(packer)
        self.amount.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ClawbackOp:
        asset = Asset.unpack(unpacker)
        from_ = MuxedAccount.unpack(unpacker)
        amount = Int64.unpack(unpacker)
        return cls(
            asset=asset,
            from_=from_,
            amount=amount,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ClawbackOp:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ClawbackOp:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.asset,
                self.from_,
                self.amount,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.asset == other.asset
            and self.from_ == other.from_
            and self.amount == other.amount
        )

    def __repr__(self):
        out = [
            f"asset={self.asset}",
            f"from_={self.from_}",
            f"amount={self.amount}",
        ]
        return f"<ClawbackOp [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["PaymentResultCode"]


class PaymentResultCode(IntEnum):
    """
    XDR Source Code::

        enum PaymentResultCode
        {
            // codes considered as "success" for the operation
            PAYMENT_SUCCESS = 0, // payment successfully completed

            // codes considered as "failure" for the operation
            PAYMENT_MALFORMED = -1,          // bad input
            PAYMENT_UNDERFUNDED = -2,        // not enough funds in source account
            PAYMENT_SRC_NO_TRUST = -3,       // no trust line on source account
            PAYMENT_SRC_NOT_AUTHORIZED = -4, // source not authorized to transfer
            PAYMENT_NO_DESTINATION = -5,     // destination account does not exist
            PAYMENT_NO_TRUST = -6,       // destination missing a trust line for asset
            PAYMENT_NOT_AUTHORIZED = -7, // destination not authorized to hold asset
            PAYMENT_LINE_FULL = -8,      // destination would go above their limit
            PAYMENT_NO_ISSUER = -9       // missing issuer on asset
        };
    """

    PAYMENT_SUCCESS = 0
    PAYMENT_MALFORMED = -1
    PAYMENT_UNDERFUNDED = -2
    PAYMENT_SRC_NO_TRUST = -3
    PAYMENT_SRC_NOT_AUTHORIZED = -4
    PAYMENT_NO_DESTINATION = -5
    PAYMENT_NO_TRUST = -6
    PAYMENT_NOT_AUTHORIZED = -7
    PAYMENT_LINE_FULL = -8
    PAYMENT_NO_ISSUER = -9

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> PaymentResultCode:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> PaymentResultCode:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> PaymentResultCode:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .constants import *
from .contract_cost_param_entry import ContractCostParamEntry

__all__ = ["ContractCostParams"]


class ContractCostParams:
    """
    XDR Source Code::

        typedef ContractCostParamEntry ContractCostParams<CONTRACT_COST_COUNT_LIMIT>;
    """

    def __init__(self, contract_cost_params: List[ContractCostParamEntry]) -> None:
        _expect_max_length = CONTRACT_COST_COUNT_LIMIT
        if contract_cost_params and len(contract_cost_params) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `contract_cost_params` should be {_expect_max_length}, but got {len(contract_cost_params)}."
            )
        self.contract_cost_params = contract_cost_params

    def pack(self, packer: Packer) -> None:
        packer.pack_uint(len(self.contract_cost_params))
        for contract_cost_params_item in self.contract_cost_params:
            contract_cost_params_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ContractCostParams:
        length = unpacker.unpack_uint()
        contract_cost_params = []
        for _ in range(length):
            contract_cost_params.append(ContractCostParamEntry.unpack(unpacker))
        return cls(contract_cost_params)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ContractCostParams:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ContractCostParams:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(self.contract_cost_params)

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.contract_cost_params == other.contract_cost_params

    def __repr__(self):
        return (
            f"<ContractCostParams [contract_cost_params={self.contract_cost_params}]>"
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .int32 import Int32

__all__ = ["Price"]


class Price:
    """
    XDR Source Code::

        struct Price
        {
            int32 n; // numerator
            int32 d; // denominator
        };
    """

    def __init__(
        self,
        n: Int32,
        d: Int32,
    ) -> None:
        self.n = n
        self.d = d

    def pack(self, packer: Packer) -> None:
        self.n.pack(packer)
        self.d.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> Price:
        n = Int32.unpack(unpacker)
        d = Int32.unpack(unpacker)
        return cls(
            n=n,
            d=d,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> Price:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> Price:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.n,
                self.d,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.n == other.n and self.d == other.d

    def __repr__(self):
        out = [
            f"n={self.n}",
            f"d={self.d}",
        ]
        return f"<Price [{', '.join(out)}]>"



from ...call_builder.base.base_call_builder import BaseCallBuilder

__all__ = ["BaseAssetsCallBuilder"]


class BaseAssetsCallBuilder(BaseCallBuilder):
    """Creates a new :class:`AssetsCallBuilder` pointed to server defined by horizon_url.

    See `List All Assets <https://developers.stellar.org/api/resources/assets/list/>`__ for more information.

    :param horizon_url: Horizon server URL.
    """

    def __init__(self, horizon_url: str) -> None:
        super().__init__(horizon_url)
        self.endpoint: str = "assets"

    def for_code(self, asset_code: str):
        """This endpoint filters all assets by the asset code.

        See `List All Assets <https://developers.stellar.org/api/resources/assets/list/>`__ for more information.

        :param asset_code: asset code, for example: `USD`
        :return: current AssetCallBuilder instance
        """
        self._add_query_param("asset_code", asset_code)
        return self

    def for_issuer(self, asset_issuer: str):
        """This endpoint filters all assets by the asset issuer.

        See `List All Assets <https://developers.stellar.org/api/resources/assets/list/>`__ for more information.

        :param asset_issuer: asset issuer,
            for example: ``"GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD"``
        :return: current AssetCallBuilder instance
        """
        self._add_query_param("asset_issuer", asset_issuer)
        return self



from . import xdr as stellar_xdr

__all__ = ["DecoratedSignature"]


class DecoratedSignature:
    def __init__(
        self,
        signature_hint: bytes,
        signature: bytes,
    ) -> None:
        """The :class:`DecoratedSignature` object, which represents a DecoratedSignature on Stellar's network.

        :param signature_hint: The signer hint
        :param signature: The signature
        """
        self.signature_hint: bytes = signature_hint
        self.signature: bytes = signature

    def to_xdr_object(self) -> stellar_xdr.DecoratedSignature:
        """Returns the xdr object for this DecoratedSignature object.

        :return: XDR DecoratedSignature object
        """
        signature_hint = stellar_xdr.SignatureHint(self.signature_hint)
        signature = stellar_xdr.Signature(self.signature)
        return stellar_xdr.DecoratedSignature(signature_hint, signature)

    @classmethod
    def from_xdr_object(
        cls, xdr_object: stellar_xdr.DecoratedSignature
    ) -> "DecoratedSignature":
        """Create a :class:`DecoratedSignature` from an XDR DecoratedSignature object.

        :param xdr_object: The XDR DecoratedSignature object.
        :return: A new :class:`DecoratedSignature` object from the given XDR DecoratedSignature object.
        """
        signature_hint = xdr_object.hint.signature_hint
        signature = xdr_object.signature.signature
        return cls(signature_hint, signature)

    def __hash__(self):
        return hash((self.signature_hint, self.signature))

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.signature_hint == other.signature_hint
            and self.signature == other.signature
        )

    def __repr__(self):
        return f"<DecoratedSignature [signature_hint={self.signature_hint}, signature={self.signature}]>"



from ...call_builder.base.base_call_builder import BaseCallBuilder

__all__ = ["BaseDataCallBuilder"]


class BaseDataCallBuilder(BaseCallBuilder):
    """Creates a new :class:`DataCallBuilder` pointed to server defined by horizon_url.

    See `Retrieve an Account's Data <https://developers.stellar.org/api/resources/accounts/data/>`__ for more information.

    :param horizon_url: Horizon server URL.
    :param account_id: account id, for example: ``"GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD"``
    :param data_name: Key name
    """

    def __init__(self, account_id: str, data_name: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.endpoint: str = f"/accounts/{account_id}/data/{data_name}"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .uint32 import Uint32

__all__ = ["SendMore"]


class SendMore:
    """
    XDR Source Code::

        struct SendMore
        {
            uint32 numMessages;
        };
    """

    def __init__(
        self,
        num_messages: Uint32,
    ) -> None:
        self.num_messages = num_messages

    def pack(self, packer: Packer) -> None:
        self.num_messages.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SendMore:
        num_messages = Uint32.unpack(unpacker)
        return cls(
            num_messages=num_messages,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SendMore:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SendMore:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash((self.num_messages,))

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.num_messages == other.num_messages

    def __repr__(self):
        out = [
            f"num_messages={self.num_messages}",
        ]
        return f"<SendMore [{', '.join(out)}]>"



from . import xdr as stellar_xdr

__all__ = ["LedgerBounds"]


class LedgerBounds:
    """LedgerBounds represents the ledger interval that a transaction is valid.

    :param min_ledger: The minimum ledger this transaction is valid at, or after.
        Cannot be negative. If the value is ``0``, the transaction is valid immediately.
    :param max_ledger: The maximum ledger this transaction is valid before.
        Cannot be negative. If the value is ``0``, the transaction is valid indefinitely.
    :raises: :exc:`ValueError <stellar_sdk.exceptions.ValueError>`: if `max_ledger` less than `min_ledger`.
    """

    def __init__(self, min_ledger: int, max_ledger: int) -> None:
        if min_ledger < 0:
            raise ValueError("min_ledger cannot be negative.")

        if max_ledger < 0:
            raise ValueError("max_ledger cannot be negative.")

        if 0 < max_ledger < min_ledger:
            raise ValueError("min_ledger cannot be greater than max_ledger.")

        self.min_ledger: int = min_ledger
        self.max_ledger: int = max_ledger

    def to_xdr_object(self) -> stellar_xdr.LedgerBounds:
        """Returns the xdr object for this LedgerBounds object.

        :return: XDR LedgerBounds object
        """
        min_ledger = stellar_xdr.Uint32(self.min_ledger)
        max_ledger = stellar_xdr.Uint32(self.max_ledger)
        return stellar_xdr.LedgerBounds(min_ledger, max_ledger)

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.LedgerBounds) -> "LedgerBounds":
        """Create a :class:`LedgerBounds` from an XDR LedgerBounds object.

        :param xdr_object: The XDR LedgerBounds object.
        :return: A new :class:`LedgerBounds` object from the given XDR LedgerBounds object.
        """
        return cls(
            min_ledger=xdr_object.min_ledger.uint32,
            max_ledger=xdr_object.max_ledger.uint32,
        )

    def __hash__(self):
        return hash((self.min_ledger, self.max_ledger))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.min_ledger == other.min_ledger and self.max_ledger == other.max_ledger
        )

    def __repr__(self):
        return f"<LedgerBounds [min_ledger={self.min_ledger}, max_ledger={self.max_ledger}]>"



from typing import Union

from . import xdr as stellar_xdr
from .keypair import Keypair
from .muxed_account import MuxedAccount
from .transaction import Transaction
from .transaction_envelope import TransactionEnvelope

__all__ = ["FeeBumpTransaction"]


class FeeBumpTransaction:
    """The :class:`FeeBumpTransaction` object, which represents a fee bump transaction
    on Stellar's network.

    See `Fee-Bump Transactions <https://developers.stellar.org/docs/glossary/fee-bumps/>`__ for more information.
    See `CAP-0015 <https://github.com/stellar/stellar-protocol/blob/master/core/cap-0015.md>`__ for more information.

    :param fee_source: The account paying for the transaction.
    :param fee: The max fee willing to pay for the transaction (**in stroops**).
    :param inner_transaction_envelope: The TransactionEnvelope to be bumped by the fee bump transaction.
    """

    def __init__(
        self,
        fee_source: Union[MuxedAccount, Keypair, str],
        fee: int,
        inner_transaction_envelope: TransactionEnvelope,
    ) -> None:
        if isinstance(fee_source, str):
            fee_source = MuxedAccount.from_account(fee_source)
        if isinstance(fee_source, Keypair):
            fee_source = MuxedAccount.from_account(fee_source.public_key)

        self.fee_source: MuxedAccount = fee_source
        self.fee: int = fee
        self.inner_transaction_envelope: TransactionEnvelope = (
            inner_transaction_envelope.to_transaction_envelope_v1()
        )
        self._inner_transaction: Transaction = (
            self.inner_transaction_envelope.transaction
        )

    def to_xdr_object(self) -> stellar_xdr.FeeBumpTransaction:
        """Get an XDR object representation of this :class:`FeeBumpTransaction`.

        :return: XDR Transaction object
        """

        fee_source = self.fee_source.to_xdr_object()
        ext = stellar_xdr.FeeBumpTransactionExt(0)
        inner_tx = stellar_xdr.FeeBumpTransactionInnerTx(
            type=stellar_xdr.EnvelopeType.ENVELOPE_TYPE_TX,
            v1=self.inner_transaction_envelope.to_xdr_object().v1,
        )
        return stellar_xdr.FeeBumpTransaction(
            fee_source=fee_source,
            fee=stellar_xdr.Int64(self.fee),
            inner_tx=inner_tx,
            ext=ext,
        )

    @classmethod
    def from_xdr_object(
        cls, xdr_object: stellar_xdr.FeeBumpTransaction, network_passphrase: str
    ) -> "FeeBumpTransaction":
        """Create a new :class:`FeeBumpTransaction` from an XDR object.

        :param xdr_object: The XDR object that represents a fee bump transaction.
        :param network_passphrase: The network to connect to for verifying and retrieving additional attributes from.

        :return: A new :class:`FeeBumpTransaction` object from the given XDR Transaction object.
        """
        source = MuxedAccount.from_xdr_object(xdr_object.fee_source)
        te = stellar_xdr.TransactionEnvelope(
            type=stellar_xdr.EnvelopeType.ENVELOPE_TYPE_TX, v1=xdr_object.inner_tx.v1
        )
        inner_transaction_envelope = TransactionEnvelope.from_xdr_object(
            te, network_passphrase
        )
        tx = cls(
            fee_source=source,
            fee=xdr_object.fee.int64,
            inner_transaction_envelope=inner_transaction_envelope,
        )
        return tx

    @classmethod
    def from_xdr(cls, xdr: str, network_passphrase: str) -> "FeeBumpTransaction":
        """Create a new :class:`FeeBumpTransaction` from an XDR string.

        :param xdr: The XDR string that represents a transaction.
        :param network_passphrase: The network to connect to for verifying and retrieving additional attributes from.

        :return: A new :class:`FeeBumpTransaction` object from the given XDR FeeBumpTransaction base64 string object.
        """
        xdr_object = stellar_xdr.FeeBumpTransaction.from_xdr(xdr)
        return cls.from_xdr_object(xdr_object, network_passphrase)

    def __hash__(self):
        return hash(
            (
                self.fee_source,
                self.fee,
                self.inner_transaction_envelope,
            )
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.fee_source == other.fee_source
            and self.fee == other.fee
            and self.inner_transaction_envelope == other.inner_transaction_envelope
        )

    def __repr__(self):
        return (
            f"<FeeBumpTransaction [fee_source={self.fee_source}, "
            f"fee={self.fee}, inner_transaction_envelope={self.inner_transaction_envelope}]>"
        )
